#!/usr/bin/python
"""
Author: Volker Schwicking, vs@heg.com

Salt tcpdumper to visualize whats happening network-wise on
the salt-master. It uses pcapy and inspects all incoming networks
packets and filters only the ones relevant to salt-communication.

based on: http://oss.coresecurity.com/projects/pcapy.html

$ salt-tcpdump.py -n 2

Will print the overall tcp-status of tcp-connections to salts
default ports in a two second interval.


$ salt-tcpdump.py -I -n 2

Will print the number of IPs making new connections to salts
default ports.


$ salt-tcpdump.py -I -n 2 -i eth1

Same as before but on eth1 instead of the default eth0.

Rough equivalents to this script could be:

For Port 4505
tcpdump "tcp[tcpflags] & tcp-syn != 0" and port 4505 and "tcp[tcpflags] & tcp-ack == 0"

For Port 4506
tcpdump "tcp[tcpflags] & tcp-syn != 0" and port 4506 and "tcp[tcpflags] & tcp-ack == 0"
"""
# pylint: disable=resource-leakage

import argparse
import socket
import sys
import time
from struct import unpack

import pcapy  # pylint: disable=import-error,3rd-party-module-not-gated


class ArgParser:
    """
    Simple Argument-Parser class
    """

    def __init__(self):
        """
        Init the Parser
        """
        self.main_parser = argparse.ArgumentParser()
        self.add_args()

    def add_args(self):
        """
        Add new arguments
        """

        self.main_parser.add_argument(
            "-i",
            type=str,
            default="eth0",
            dest="iface",
            required=False,
            help="The interface to dump the master runs on(default:eth0)",
        )

        self.main_parser.add_argument(
            "-n",
            type=int,
            default=5,
            dest="ival",
            required=False,
            help="Interval for printing stats (default:5)",
        )

        self.main_parser.add_argument(
            "-I",
            type=bool,
            default=False,
            const=True,
            nargs="?",
            dest="only_ip",
            required=False,
            help="Print unique IPs making new connections with SYN set",
        )

    def parse_args(self):
        """
        parses and returns the given arguments in a namespace object
        """
        return self.main_parser.parse_args()


class PCAPParser:
    """
    parses a network packet on given device and
    returns source, target, source_port and dest_port
    """

    def __init__(self, iface):
        self.iface = iface

    def run(self):
        """
        main loop for the packet-parser
        """
        # open device
        # Arguments here are:
        #   device
        #   snaplen (maximum number of bytes to capture _per_packet_)
        #   promiscious mode (1 for true)
        #   timeout (in milliseconds)
        cap = pcapy.open_live(self.iface, 65536, 1, 0)

        count = 0
        l_time = None

        while 1:

            packet_data = {"ip": {}, "tcp": {}}

            (header, packet) = cap.next()

            eth_length, eth_protocol = self.parse_ether(packet)

            # Parse IP packets, IP Protocol number = 8
            if eth_protocol == 8:
                # Parse IP header
                # take first 20 characters for the ip header
                (
                    version_ihl,
                    version,
                    ihl,
                    iph_length,
                    ttl,
                    protocol,
                    s_addr,
                    d_addr,
                ) = self.parse_ip(packet, eth_length)
                packet_data["ip"]["s_addr"] = s_addr
                packet_data["ip"]["d_addr"] = d_addr

                # TCP protocol
                if protocol == 6:

                    source_port, dest_port, flags, data = self.parse_tcp(
                        packet, iph_length, eth_length
                    )
                    packet_data["tcp"]["d_port"] = dest_port
                    packet_data["tcp"]["s_port"] = source_port
                    packet_data["tcp"]["flags"] = flags
                    packet_data["tcp"]["data"] = data
                    yield packet_data

    def parse_ether(self, packet):
        """
        parse ethernet_header and return size and protocol
        """
        eth_length = 14

        eth_header = packet[:eth_length]
        eth = unpack("!6s6sH", eth_header)
        eth_protocol = socket.ntohs(eth[2])
        return eth_length, eth_protocol

    def parse_ip(self, packet, eth_length):
        """
        parse ip_header and return all ip data fields
        """
        # Parse IP header
        # take first 20 characters for the ip header
        ip_header = packet[eth_length : 20 + eth_length]

        # now unpack them:)
        iph = unpack("!BBHHHBBH4s4s", ip_header)

        version_ihl = iph[0]
        version = version_ihl >> 4
        ihl = version_ihl & 0xF

        iph_length = ihl * 4

        ttl = iph[5]
        protocol = iph[6]
        s_addr = socket.inet_ntoa(iph[8])
        d_addr = socket.inet_ntoa(iph[9])

        return [version_ihl, version, ihl, iph_length, ttl, protocol, s_addr, d_addr]

    def parse_tcp(self, packet, iph_length, eth_length):
        """
        parse tcp_data and return source_port,
        dest_port and actual packet data
        """
        p_len = iph_length + eth_length
        tcp_header = packet[p_len : p_len + 20]

        # now unpack them:)
        tcph = unpack("!H HLLBBHHH", tcp_header)
        #  H     H     L   L   B   B      H   H   H
        #  2b    2b    4b  4b  1b  1b     2b  2b  2b
        #  sport dport seq ack res flags  win chk up
        # (22,   36513, 3701969065, 2346113113, 128, 24, 330, 33745, 0)
        source_port = tcph[0]

        dest_port = tcph[1]
        sequence = tcph[2]
        acknowledgement = tcph[3]
        doff_reserved = tcph[4]
        tcph_length = doff_reserved >> 4
        tcp_flags = tcph[5]

        h_size = eth_length + iph_length + tcph_length * 4
        data_size = len(packet) - h_size

        data = packet[h_size:]

        return source_port, dest_port, tcp_flags, data


class SaltNetstat:
    """
    Reads /proc/net/tcp and returns all connections
    """

    def proc_tcp(self):
        """
        Read the table of tcp connections & remove header
        """
        with open("/proc/net/tcp", encoding="utf-8") as tcp_f:
            content = tcp_f.readlines()
            content.pop(0)
        return content

    def hex2dec(self, hex_s):
        """
        convert hex to dezimal
        """
        return str(int(hex_s, 16))

    def ip(self, hex_s):
        """
        convert into readable ip
        """
        ip = [
            (self.hex2dec(hex_s[6:8])),
            (self.hex2dec(hex_s[4:6])),
            (self.hex2dec(hex_s[2:4])),
            (self.hex2dec(hex_s[0:2])),
        ]
        return ".".join(ip)

    def remove_empty(self, array):
        """
        create new list without empty entries
        """
        return [x for x in array if x != ""]

    def convert_ip_port(self, array):
        """
        hex_ip:hex_port to str_ip:str_port
        """
        host, port = array.split(":")
        return self.ip(host), self.hex2dec(port)

    def run(self):
        """
        main loop for netstat
        """
        while 1:
            ips = {"ips/4505": {}, "ips/4506": {}}
            content = self.proc_tcp()

            for line in content:
                line_array = self.remove_empty(line.split(" "))
                l_host, l_port = self.convert_ip_port(line_array[1])
                r_host, r_port = self.convert_ip_port(line_array[2])
                if l_port == "4505":
                    if r_host not in ips["ips/4505"]:
                        ips["ips/4505"][r_host] = 0
                    ips["ips/4505"][r_host] += 1
                if l_port == "4506":
                    if r_host not in ips["ips/4506"]:
                        ips["ips/4506"][r_host] = 0
                    ips["ips/4506"][r_host] += 1

            yield (len(ips["ips/4505"]), len(ips["ips/4506"]))
            time.sleep(0.5)


def filter_new_cons(packet):
    """
    filter packets by there tcp-state and
    returns codes for specific states
    """
    flags = []
    TCP_FIN = 0x01
    TCP_SYN = 0x02
    TCP_RST = 0x04
    TCP_PSH = 0x08
    TCP_ACK = 0x10
    TCP_URG = 0x20
    TCP_ECE = 0x40
    TCP_CWK = 0x80

    if packet["tcp"]["flags"] & TCP_FIN:
        flags.append("FIN")
    elif packet["tcp"]["flags"] & TCP_SYN:
        flags.append("SYN")
    elif packet["tcp"]["flags"] & TCP_RST:
        flags.append("RST")
    elif packet["tcp"]["flags"] & TCP_PSH:
        flags.append("PSH")
    elif packet["tcp"]["flags"] & TCP_ACK:
        flags.append("ACK")
    elif packet["tcp"]["flags"] & TCP_URG:
        flags.append("URG")
    elif packet["tcp"]["flags"] & TCP_ECE:
        flags.append("ECE")
    elif packet["tcp"]["flags"] & TCP_CWK:
        flags.append("CWK")
    else:
        print("UNKNOWN PACKET")

    if packet["tcp"]["d_port"] == 4505:
        # track new connections
        if "SYN" in flags and len(flags) == 1:
            return 10
        # track closing connections
        elif "FIN" in flags:
            return 12

    elif packet["tcp"]["d_port"] == 4506:
        # track new connections
        if "SYN" in flags and len(flags) == 1:
            return 100
        # track closing connections
        elif "FIN" in flags:
            return 120
    # packet does not match requirements
    else:
        return None


def main():
    """
    main loop for whole script
    """
    # passed parameters
    args = vars(ArgParser().parse_args())

    # reference timer for printing in intervals
    r_time = 0

    # the ports we want to monitor
    ports = [4505, 4506]

    print("Sniffing device {}".format(args["iface"]))

    stat = {
        "4506/new": 0,
        "4506/est": 0,
        "4506/fin": 0,
        "4505/new": 0,
        "4505/est": 0,
        "4505/fin": 0,
        "ips/4505": 0,
        "ips/4506": 0,
    }

    if args["only_ip"]:
        print(
            "IPs making new connections (ports:{}, interval:{})".format(
                ports, args["ival"]
            )
        )
    else:
        print(
            "Salt-Master Network Status (ports:{}, interval:{})".format(
                ports, args["ival"]
            )
        )
    try:
        while 1:
            s_time = int(time.time())

            packet = next(PCAPParser(args["iface"]).run())

            p_state = filter_new_cons(packet)

            ips_auth = []
            ips_push = []

            # new connection to 4505
            if p_state == 10:
                stat["4505/new"] += 1
                if packet["ip"]["s_addr"] not in ips_auth:
                    ips_auth.append(packet["ip"]["s_addr"])
            # closing connection to 4505
            elif p_state == 12:
                stat["4505/fin"] += 1

            # new connection to 4506
            elif p_state == 100:
                stat["4506/new"] += 1
                if packet["ip"]["s_addr"] not in ips_push:
                    ips_push.append(packet["ip"]["s_addr"])
            # closing connection to 4506
            elif p_state == 120:
                stat["4506/fin"] += 1

            # get the established connections to 4505 and 4506
            # these would only show up in tcpdump if data is transferred
            # but then with different flags (PSH, etc.)
            stat["4505/est"], stat["4506/est"] = next(SaltNetstat().run())

            # print only in intervals
            if (s_time % args["ival"]) == 0:
                # prevent printing within the same second
                if r_time != s_time:
                    if args["only_ip"]:
                        msg = "IPs/4505: {}, IPs/4506: {}".format(
                            len(ips_auth), len(ips_push)
                        )
                    else:
                        msg = "4505=>[ est: {}, ".format(stat["4505/est"])
                        msg += "new: {}/s, ".format(stat["4505/new"] / args["ival"])
                        msg += "fin: {}/s ] ".format(stat["4505/fin"] / args["ival"])

                        msg += " 4506=>[ est: {}, ".format(stat["4506/est"])
                        msg += "new: {}/s, ".format(stat["4506/new"] / args["ival"])
                        msg += "fin: {}/s ]".format(stat["4506/fin"] / args["ival"])

                    print(msg)

                    # reset the so far collected stats
                    for item in stat:
                        stat[item] = 0
                    r_time = s_time

    except KeyboardInterrupt:
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
Manage the information in the aliases file
"""

import os
import re
import stat
import tempfile

import salt.utils.atomicfile
import salt.utils.files
import salt.utils.path
import salt.utils.stringutils
from salt.exceptions import SaltInvocationError

__outputter__ = {
    "rm_alias": "txt",
    "has_target": "txt",
    "get_target": "txt",
    "set_target": "txt",
    "list_aliases": "yaml",
}

__ALIAS_RE = re.compile(r"([^:#]*)\s*:?\s*([^#]*?)(\s+#.*|$)")


def __get_aliases_filename():
    """
    Return the path to the appropriate aliases file
    """
    return os.path.realpath(__salt__["config.option"]("aliases.file"))


def __parse_aliases():
    """
    Parse the aliases file, and return a list of line components:

    [
      (alias1, target1, comment1),
      (alias2, target2, comment2),
    ]
    """
    afn = __get_aliases_filename()
    ret = []
    if not os.path.isfile(afn):
        return ret
    with salt.utils.files.fopen(afn, "r") as ifile:
        for line in ifile:
            line = salt.utils.stringutils.to_unicode(line)
            match = __ALIAS_RE.match(line)
            if match:
                ret.append(match.groups())
            else:
                ret.append((None, None, line.strip()))
    return ret


def __write_aliases_file(lines):
    """
    Write a new copy of the aliases file.  Lines is a list of lines
    as returned by __parse_aliases.
    """
    afn = __get_aliases_filename()
    adir = os.path.dirname(afn)

    with tempfile.NamedTemporaryFile(dir=adir, delete=False) as out:

        if not __opts__.get("integration.test", False):
            if os.path.isfile(afn):
                afn_st = os.stat(afn)
                os.chmod(out.name, stat.S_IMODE(afn_st.st_mode))
                os.chown(out.name, afn_st.st_uid, afn_st.st_gid)
            else:
                os.chmod(out.name, 0o644)
                os.chown(out.name, 0, 0)

        for line_alias, line_target, line_comment in lines:
            if isinstance(line_target, list):
                line_target = ", ".join(line_target)
            if not line_comment:
                line_comment = ""
            if line_alias and line_target:
                write_line = f"{line_alias}: {line_target}{line_comment}\n"
            else:
                write_line = f"{line_comment}\n"
            write_line = write_line.encode(__salt_system_encoding__)
            out.write(write_line)

    salt.utils.atomicfile.atomic_rename(out.name, afn)

    # Search $PATH for the newalises command
    newaliases = salt.utils.path.which("newaliases")
    if newaliases is not None:
        __salt__["cmd.run"](newaliases)

    return True


def list_aliases():
    """
    Return the aliases found in the aliases file in this format::

        {'alias': 'target'}

    CLI Example:

    .. code-block:: bash

        salt '*' aliases.list_aliases
    """
    ret = {alias: target for (alias, target, _) in __parse_aliases() if alias}
    return ret


def get_target(alias):
    """
    Return the target associated with an alias

    CLI Example:

    .. code-block:: bash

        salt '*' aliases.get_target alias
    """
    aliases = list_aliases()
    if alias in aliases:
        return aliases[alias]
    return ""


def has_target(alias, target):
    """
    Return true if the alias/target is set

    CLI Example:

    .. code-block:: bash

        salt '*' aliases.has_target alias target
    """
    if target == "":
        raise SaltInvocationError("target can not be an empty string")
    aliases = list_aliases()
    if alias not in aliases:
        return False
    if isinstance(target, list):
        target = ", ".join(target)
    return target == aliases[alias]


def set_target(alias, target):
    """
    Set the entry in the aliases file for the given alias, this will overwrite
    any previous entry for the given alias or create a new one if it does not
    exist.

    CLI Example:

    .. code-block:: bash

        salt '*' aliases.set_target alias target
    """

    if alias == "":
        raise SaltInvocationError("alias can not be an empty string")

    if target == "":
        raise SaltInvocationError("target can not be an empty string")

    if get_target(alias) == target:
        return True

    lines = __parse_aliases()
    out = []
    ovr = False
    for line_alias, line_target, line_comment in lines:
        if line_alias == alias:
            if not ovr:
                out.append((alias, target, line_comment))
                ovr = True
        else:
            out.append((line_alias, line_target, line_comment))
    if not ovr:
        out.append((alias, target, ""))

    __write_aliases_file(out)
    return True


def rm_alias(alias):
    """
    Remove an entry from the aliases file

    CLI Example:

    .. code-block:: bash

        salt '*' aliases.rm_alias alias
    """
    if not get_target(alias):
        return True

    lines = __parse_aliases()
    out = []
    for line_alias, line_target, line_comment in lines:
        if line_alias != alias:
            out.append((line_alias, line_target, line_comment))

    __write_aliases_file(out)
    return True



"""
State module to manage Elasticsearch index templates

.. versionadded:: 2015.8.0
.. deprecated:: 2017.7.0
   Use elasticsearch state instead
"""

import logging

log = logging.getLogger(__name__)


def absent(name):
    """
    Ensure that the named index template is absent.

    name
        Name of the index to remove
    """

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    try:
        index_template = __salt__["elasticsearch.index_template_get"](name=name)
        if index_template and name in index_template:
            if __opts__["test"]:
                ret["comment"] = f"Index template {name} will be removed"
                ret["changes"]["old"] = index_template[name]
                ret["result"] = None
            else:
                ret["result"] = __salt__["elasticsearch.index_template_delete"](
                    name=name
                )
                if ret["result"]:
                    ret["comment"] = "Successfully removed index template {}".format(
                        name
                    )
                    ret["changes"]["old"] = index_template[name]
                else:
                    ret["comment"] = (
                        "Failed to remove index template {} for unknown reasons".format(
                            name
                        )
                    )
        else:
            ret["comment"] = f"Index template {name} is already absent"
    except Exception as err:  # pylint: disable=broad-except
        ret["result"] = False
        ret["comment"] = str(err)

    return ret


def present(name, definition):
    """
    .. versionadded:: 2015.8.0
    .. versionchanged:: 2017.3.0
        Marked ``definition`` as required.

    Ensure that the named index templat eis present.

    name
        Name of the index to add

    definition
        Required dict for creation parameters as per https://www.elastic.co/guide/en/elasticsearch/reference/current/indices-templates.html

    **Example:**

    .. code-block:: yaml

        mytestindex2_template:
          elasticsearch_index_template.present:
            - definition:
                template: logstash-*
                order: 1
                settings:
                  number_of_shards: 1
    """

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    try:
        index_template_exists = __salt__["elasticsearch.index_template_exists"](
            name=name
        )
        if not index_template_exists:
            if __opts__["test"]:
                ret["comment"] = (
                    f"Index template {name} does not exist and will be created"
                )
                ret["changes"] = {"new": definition}
                ret["result"] = None
            else:
                output = __salt__["elasticsearch.index_template_create"](
                    name=name, body=definition
                )
                if output:
                    ret["comment"] = "Successfully created index template {}".format(
                        name
                    )
                    ret["changes"] = {
                        "new": __salt__["elasticsearch.index_template_get"](name=name)[
                            name
                        ]
                    }
                else:
                    ret["result"] = False
                    ret["comment"] = "Cannot create index template {}, {}".format(
                        name, output
                    )
        else:
            ret["comment"] = f"Index template {name} is already present"
    except Exception as err:  # pylint: disable=broad-except
        ret["result"] = False
        ret["comment"] = str(err)

    return ret



"""
Installation of Cygwin packages.

A state module to manage cygwin packages. Packages can be installed
or removed.

.. code-block:: yaml

    dos2unix:
      cyg.installed
"""

import logging

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if cyg module is available in __salt__.
    """
    if "cyg.list" in __salt__:
        return True
    return (False, "cyg module could not be loaded")


def installed(name, cyg_arch="x86_64", mirrors=None):
    """
    Make sure that a package is installed.

    name
        The name of the package to install

    cyg_arch : x86_64
        The cygwin architecture to install the package into.
        Current options are x86 and x86_64

    mirrors : None
        List of mirrors to check.
        None will use a default mirror (kernel.org)

    CLI Example:

    .. code-block:: yaml

        rsync:
          cyg.installed:
            - mirrors:
              - http://mirror/without/public/key: ""
              - http://mirror/with/public/key: http://url/of/public/key
    """
    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    if cyg_arch not in ["x86", "x86_64"]:
        ret["result"] = False
        ret["comment"] = "The 'cyg_arch' argument must be one of 'x86' or 'x86_64'"
        return ret

    log.debug("Installed State: Initial Mirror list: %s", mirrors)

    if not __salt__["cyg.check_valid_package"](
        name, cyg_arch=cyg_arch, mirrors=mirrors
    ):
        ret["result"] = False
        ret["comment"] = "Invalid package name."
        return ret

    pkgs = __salt__["cyg.list"](name, cyg_arch)
    if name in pkgs:
        ret["result"] = True
        ret["comment"] = "Package is already installed."
        return ret

    if __opts__["test"]:
        ret["comment"] = f"The package {name} would have been installed"
        return ret

    if __salt__["cyg.install"](name, cyg_arch=cyg_arch, mirrors=mirrors):
        ret["result"] = True
        ret["changes"][name] = "Installed"
        ret["comment"] = "Package was successfully installed"
    else:
        ret["result"] = False
        ret["comment"] = "Could not install package."

    return ret


def removed(name, cyg_arch="x86_64", mirrors=None):
    """
    Make sure that a package is not installed.

    name
        The name of the package to uninstall

    cyg_arch : x86_64
        The cygwin architecture to remove the package from.
        Current options are x86 and x86_64

    mirrors : None
        List of mirrors to check.
        None will use a default mirror (kernel.org)

    CLI Example:

    .. code-block:: yaml

        rsync:
          cyg.removed:
            - mirrors:
              - http://mirror/without/public/key: ""
              - http://mirror/with/public/key: http://url/of/public/key
    """
    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    if cyg_arch not in ["x86", "x86_64"]:
        ret["result"] = False
        ret["comment"] = "The 'cyg_arch' argument must be one of 'x86' or 'x86_64'"
        return ret

    if not __salt__["cyg.check_valid_package"](
        name, cyg_arch=cyg_arch, mirrors=mirrors
    ):
        ret["result"] = False
        ret["comment"] = "Invalid package name."
        return ret

    if name not in __salt__["cyg.list"](name, cyg_arch):
        ret["result"] = True
        ret["comment"] = "Package is not installed."
        return ret

    if __opts__["test"]:
        ret["comment"] = f"The package {name} would have been removed"
        return ret
    if __salt__["cyg.uninstall"](name, cyg_arch):
        ret["result"] = True
        ret["changes"][name] = "Removed"
        ret["comment"] = "Package was successfully removed."
    else:
        ret["result"] = False
        ret["comment"] = "Could not remove package."
    return ret


def updated(name=None, cyg_arch="x86_64", mirrors=None):
    """
    Make sure all packages are up to date.

    name : None
        No affect, salt fails poorly without the arg available

    cyg_arch : x86_64
        The cygwin architecture to update.
        Current options are x86 and x86_64

    mirrors : None
        List of mirrors to check.
        None will use a default mirror (kernel.org)

    CLI Example:

    .. code-block:: yaml

        rsync:
          cyg.updated:
            - mirrors:
              - http://mirror/without/public/key: ""
              - http://mirror/with/public/key: http://url/of/public/key
    """
    ret = {"name": "cyg.updated", "result": None, "comment": "", "changes": {}}

    if cyg_arch not in ["x86", "x86_64"]:
        ret["result"] = False
        ret["comment"] = "The 'cyg_arch' argument must be one of 'x86' or 'x86_64'"
        return ret

    if __opts__["test"]:
        ret["comment"] = "All packages would have been updated"
        return ret

    if not mirrors:
        log.warning("No mirror given, using the default.")

    before = __salt__["cyg.list"](cyg_arch=cyg_arch)
    if __salt__["cyg.update"](cyg_arch, mirrors=mirrors):
        after = __salt__["cyg.list"](cyg_arch=cyg_arch)
        differ = DictDiffer(after, before)
        ret["result"] = True
        if differ.same():
            ret["comment"] = "Nothing to update."
        else:
            ret["changes"]["added"] = list(differ.added())
            ret["changes"]["removed"] = list(differ.removed())
            ret["changes"]["changed"] = list(differ.changed())
            ret["comment"] = "All packages successfully updated."
    else:
        ret["result"] = False
        ret["comment"] = "Could not update packages."
    return ret


# https://github.com/hughdbrown/dictdiffer
# DictDiffer is licensed as MIT code
# A dictionary difference calculator
# Originally posted as:
# http://stackoverflow.com/a/1165552


class DictDiffer:
    """
    Calculate the difference between two dictionaries.

    (1) items added
    (2) items removed
    (3) keys same in both but changed values
    (4) keys same in both and unchanged values
    """

    def __init__(self, current_dict, past_dict):
        """
        Iitialize the differ.
        """
        self.current_dict, self.past_dict = current_dict, past_dict
        self.current_keys, self.past_keys = (
            set(d.keys()) for d in (current_dict, past_dict)
        )
        self.intersect = self.current_keys.intersection(self.past_keys)

    def same(self):
        """
        True if the two dicts are the same.
        """
        return self.current_dict == self.past_dict

    def added(self):
        """
        Return a set of additions to past_dict.
        """
        return self.current_keys - self.intersect

    def removed(self):
        """
        Return a set of things removed from past_dict.
        """
        return self.past_keys - self.intersect

    def changed(self):
        """
        Return a set of the keys with changed values.
        """
        return {o for o in self.intersect if self.past_dict[o] != self.current_dict[o]}

    def unchanged(self):
        """
        Return a set of the keys with unchanged values.
        """
        return {o for o in self.intersect if self.past_dict[o] == self.current_dict[o]}



"""
Management of user groups
=========================

The group module is used to create and manage group settings, groups can be
either present or absent. User/Group names can be passed to the ``adduser``,
``deluser``, and ``members`` parameters. ``adduser`` and ``deluser`` can be used
together but not with ``members``.

In Windows, if no domain is specified in the user or group name (i.e.
``DOMAIN\\username``) the module will assume a local user or group.

.. code-block:: yaml

    cheese:
      group.present:
        - gid: 7648
        - system: True
        - addusers:
          - user1
          - users2
        - delusers:
          - foo

    cheese:
      group.present:
        - gid: 7648
        - system: True
        - members:
          - foo
          - bar
          - user1
          - user2
"""

import sys

import salt.utils.platform
import salt.utils.win_functions


def _get_root_args(local):
    """
    Retrieve args to use for group.info calls depending on platform and the local flag
    """
    if not local or salt.utils.platform.is_windows():
        return {}
    return {"root": "/"}


def _changes(name, gid=None, addusers=None, delusers=None, members=None, local=False):
    """
    Return a dict of the changes required for a group if the group is present,
    otherwise return False.
    """
    lgrp = __salt__["group.info"](name, **_get_root_args(local))
    if not lgrp:
        return False

    # User and Domain names are not case sensitive in Windows. Let's make them
    # all lower case so we can compare properly
    if salt.utils.platform.is_windows():
        if lgrp["members"]:
            lgrp["members"] = [user.lower() for user in lgrp["members"]]
        if members:
            members = [
                salt.utils.win_functions.get_sam_name(user).lower() for user in members
            ]
        if addusers:
            addusers = [
                salt.utils.win_functions.get_sam_name(user).lower() for user in addusers
            ]
        if delusers:
            delusers = [
                salt.utils.win_functions.get_sam_name(user).lower() for user in delusers
            ]

    change = {}
    ret = {}
    if gid:
        try:
            gid = int(gid)
            if lgrp["gid"] != gid:
                change["gid"] = gid
        except (TypeError, ValueError):
            ret["result"] = False
            ret["comment"] = "Invalid gid"
            return ret

    if members is not None and not members:
        if set(lgrp["members"]).symmetric_difference(members):
            change["delusers"] = set(lgrp["members"])
    elif members:
        # if new member list if different than the current
        if set(lgrp["members"]).symmetric_difference(members):
            change["members"] = members

    if addusers:
        users_2add = [user for user in addusers if user not in lgrp["members"]]
        if users_2add:
            change["addusers"] = users_2add

    if delusers:
        users_2del = [user for user in delusers if user in lgrp["members"]]
        if users_2del:
            change["delusers"] = users_2del

    return change


def present(
    name,
    gid=None,
    system=False,
    addusers=None,
    delusers=None,
    members=None,
    non_unique=False,
    local=False,
):
    r"""
    .. versionchanged:: 3006.0

    Ensure that a group is present

    Args:

        name (str):
            The name of the group to manage

        gid (str):
            The group id to assign to the named group; if left empty, then the
            next available group id will be assigned. Ignored on Windows

        system (bool):
            Whether or not the named group is a system group.  This is essentially
            the '-r' option of 'groupadd'. Ignored on Windows

        addusers (list):
            List of additional users to be added as a group members. Cannot
            conflict with names in delusers. Cannot be used in conjunction with
            members.

        delusers (list):
            Ensure these user are removed from the group membership. Cannot
            conflict with names in addusers. Cannot be used in conjunction with
            members.

        members (list):
            Replace existing group members with a list of new members. Cannot be
            used in conjunction with addusers or delusers.

        non_unique (bool):
            Allow creating groups with duplicate (non-unique) GIDs

            .. versionadded:: 3006.0

        local (Only on systems with lgroupadd available):
            Create the group account locally ignoring global account management
            (default is False).

            .. versionadded:: 3007.0

    Example:

    .. code-block:: yaml

        # Adds DOMAIN\db_admins and Administrators to the local db_admin group
        # Removes Users
        db_admin:
          group.present:
            - addusers:
              - DOMAIN\db_admins
              - Administrators
            - delusers:
              - Users

        # Ensures only DOMAIN\domain_admins and the local Administrator are
        # members of the local Administrators group. All other users are
        # removed
        Administrators:
          group.present:
            - members:
              - DOMAIN\domain_admins
              - Administrator
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f"Group {name} is present and up to date",
    }

    if members is not None and (addusers is not None or delusers is not None):
        ret["result"] = None
        ret["comment"] = (
            'Error: Conflicting options "members" with "addusers" and/or'
            ' "delusers" can not be used together. '
        )
        return ret

    if addusers and delusers:
        # -- if trying to add and delete the same user(s) at the same time.
        if not set(addusers).isdisjoint(set(delusers)):
            ret["result"] = None
            ret["comment"] = (
                "Error. Same user(s) can not be added and deleted simultaneously"
            )
            return ret

    changes = _changes(name, gid, addusers, delusers, members, local=local)
    if changes:
        ret["comment"] = "The following group attributes are set to be changed:\n"
        for key, val in changes.items():
            ret["comment"] += f"{key}: {val}\n"

        if __opts__["test"]:
            ret["result"] = None
            return ret

        for key, val in changes.items():
            if key == "gid":
                __salt__["group.chgid"](name, gid, non_unique=non_unique)
                continue
            if key == "addusers":
                for user in val:
                    __salt__["group.adduser"](name, user)
                continue
            if key == "delusers":
                for user in val:
                    __salt__["group.deluser"](name, user)
                continue
            if key == "members":
                __salt__["group.members"](name, ",".join(members))
                continue
        # Clear cached group data
        sys.modules[__salt__["test.ping"].__module__].__context__.pop(
            "group.getent", None
        )
        changes = _changes(name, gid, addusers, delusers, members, local=local)
        if changes:
            ret["result"] = False
            ret["comment"] += "Some changes could not be applied"
            ret["changes"] = {"Failed": changes}
        else:
            ret["changes"] = {"Final": "All changes applied successfully"}

    if changes is False:
        # The group is not present, make it!
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Group {name} set to be added"
            return ret

        grps = __salt__["group.getent"]()
        # Test if gid is free
        if gid is not None and not non_unique:
            gid_group = None
            for lgrp in grps:
                if lgrp["gid"] == gid:
                    gid_group = lgrp["name"]
                    break

            if gid_group is not None:
                ret["result"] = False
                ret["comment"] = (
                    "Group {} is not present but gid {} is already taken by"
                    " group {}".format(name, gid, gid_group)
                )
                return ret

        # Group is not present, make it.
        if salt.utils.platform.is_windows():
            add_args = {}
        else:
            add_args = {"local": local}
        if __salt__["group.add"](
            name,
            gid=gid,
            system=system,
            non_unique=non_unique,
            **add_args,
        ):
            # if members to be added
            grp_members = None
            if members:
                grp_members = ",".join(members)
            if addusers:
                grp_members = ",".join(addusers)
            if grp_members:
                __salt__["group.members"](name, grp_members)
            # Clear cached group data
            sys.modules[__salt__["test.ping"].__module__].__context__.pop(
                "group.getent", None
            )
            ret["comment"] = f"New group {name} created"
            ret["changes"] = __salt__["group.info"](name, **_get_root_args(local))
            changes = _changes(name, gid, addusers, delusers, members, local=local)
            if changes:
                ret["result"] = False
                ret["comment"] = (
                    "Group {} has been created but, some changes could not"
                    " be applied".format(name)
                )
                ret["changes"] = {"Failed": changes}
        else:
            ret["result"] = False
            ret["comment"] = f"Failed to create new group {name}"
    return ret


def absent(name, local=False):
    """
    Ensure that the named group is absent

    Args:
        name (str):
            The name of the group to remove

        local (Only on systems with lgroupdel available):

            Ensure the group account is removed locally ignoring global
            account management (default is False).

            .. versionadded:: 3007.0

    Example:

    .. code-block:: yaml

        # Removes the local group `db_admin`
        db_admin:
          group.absent
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    grp_info = __salt__["group.info"](name, **_get_root_args(local))
    if grp_info:
        # Group already exists. Remove the group.
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Group {name} is set for removal"
            return ret
        if salt.utils.platform.is_windows():
            del_args = {}
        else:
            del_args = {"local": local}
        ret["result"] = __salt__["group.delete"](name, **del_args)
        if ret["result"]:
            ret["changes"] = {name: ""}
            ret["comment"] = f"Removed group {name}"
            return ret
        else:
            ret["comment"] = f"Failed to remove group {name}"
            return ret
    else:
        ret["comment"] = "Group not present"
        return ret



import os

import pytest

import salt.modules.systemd_service as systemd
import salt.utils.systemd
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture()
def systemctl_status():
    return {
        "sshd.service": {
            "stdout": """\
* sshd.service - OpenSSH Daemon
   Loaded: loaded (/usr/lib/systemd/system/sshd.service; disabled; vendor preset: disabled)
   Active: inactive (dead)""",
            "stderr": "",
            "retcode": 3,
            "pid": 12345,
        },
        "foo.service": {
            "stdout": """\
* foo.service
   Loaded: not-found (Reason: No such file or directory)
   Active: inactive (dead)""",
            "stderr": "",
            "retcode": 3,
            "pid": 12345,
        },
    }


# This reflects systemd >= 231 behavior
@pytest.fixture()
def systemctl_status_gte_231():
    return {
        "bar.service": {
            "stdout": "Unit bar.service could not be found.",
            "stderr": "",
            "retcode": 4,
            "pid": 12345,
        },
    }


@pytest.fixture()
def list_unit_files():
    return """\
service1.service                           enabled              -
service2.service                           disabled             -
service3.service                           static               -
timer1.timer                               enabled              -
timer2.timer                               disabled             -
timer3.timer                               static               -
service4.service                           enabled              enabled
service5.service                           disabled             enabled
service6.service                           static               enabled
timer4.timer                               enabled              enabled
timer5.timer                               disabled             enabled
timer6.timer                               static               enabled
service7.service                           enabled              disabled
service8.service                           disabled             disabled
service9.service                           static               disabled
timer7.timer                               enabled              disabled
timer8.timer                               disabled             disabled
timer9.timer                               static               disabled
service10.service                          enabled
service11.service                          disabled
service12.service                          static
timer10.timer                              enabled
timer11.timer                              disabled
timer12.timer                              static"""


@pytest.fixture()
def configure_loader_modules():
    return {systemd: {}}


def test_systemctl_reload():
    """
    Test to Reloads systemctl
    """
    mock = MagicMock(
        side_effect=[
            {"stdout": "Who knows why?", "stderr": "", "retcode": 1, "pid": 12345},
            {"stdout": "", "stderr": "", "retcode": 0, "pid": 54321},
        ]
    )
    with patch.dict(systemd.__salt__, {"cmd.run_all": mock}):
        with pytest.raises(
            CommandExecutionError,
            match="Problem performing systemctl daemon-reload: Who knows why?",
        ):
            systemd.systemctl_reload()
        assert systemd.systemctl_reload() is True


def test_get_enabled(list_unit_files, systemctl_status):
    """
    Test to return a list of all enabled services
    """
    cmd_mock = MagicMock(return_value=list_unit_files)
    listdir_mock = MagicMock(return_value=["foo", "bar", "baz", "README"])
    sd_mock = MagicMock(
        return_value={x.replace(".service", "") for x in systemctl_status}
    )
    access_mock = MagicMock(
        side_effect=lambda x, y: x != os.path.join(systemd.INITSCRIPT_PATH, "README")
    )
    sysv_enabled_mock = MagicMock(side_effect=lambda x, _: x == "baz")

    with patch.dict(systemd.__salt__, {"cmd.run": cmd_mock}):
        with patch.object(os, "listdir", listdir_mock):
            with patch.object(systemd, "_get_systemd_services", sd_mock):
                with patch.object(os, "access", side_effect=access_mock):
                    with patch.object(systemd, "_sysv_enabled", sysv_enabled_mock):
                        assert systemd.get_enabled() == [
                            "baz",
                            "service1",
                            "service10",
                            "service4",
                            "service7",
                            "timer1.timer",
                            "timer10.timer",
                            "timer4.timer",
                            "timer7.timer",
                        ]


def test_get_disabled(list_unit_files, systemctl_status):
    """
    Test to return a list of all disabled services
    """
    cmd_mock = MagicMock(return_value=list_unit_files)
    # 'foo' should collide with the systemd services (as returned by
    # sd_mock) and thus not be returned by _get_sysv_services(). It doesn't
    # matter that it's not part of the _LIST_UNIT_FILES output, we just
    # want to ensure that 'foo' isn't identified as a disabled initscript
    # even though below we are mocking it to show as not enabled (since
    # only 'baz' will be considered an enabled sysv service).
    listdir_mock = MagicMock(return_value=["foo", "bar", "baz", "README"])
    sd_mock = MagicMock(
        return_value={x.replace(".service", "") for x in systemctl_status}
    )
    access_mock = MagicMock(
        side_effect=lambda x, y: x != os.path.join(systemd.INITSCRIPT_PATH, "README")
    )
    sysv_enabled_mock = MagicMock(side_effect=lambda x, _: x == "baz")

    with patch.dict(systemd.__salt__, {"cmd.run": cmd_mock}):
        with patch.object(os, "listdir", listdir_mock):
            with patch.object(systemd, "_get_systemd_services", sd_mock):
                with patch.object(os, "access", side_effect=access_mock):
                    with patch.object(systemd, "_sysv_enabled", sysv_enabled_mock):
                        assert systemd.get_disabled() == [
                            "bar",
                            "service11",
                            "service2",
                            "service5",
                            "service8",
                            "timer11.timer",
                            "timer2.timer",
                            "timer5.timer",
                            "timer8.timer",
                        ]


def test_get_static(list_unit_files, systemctl_status):
    """
    Test to return a list of all disabled services
    """
    cmd_mock = MagicMock(return_value=list_unit_files)
    # 'foo' should collide with the systemd services (as returned by
    # sd_mock) and thus not be returned by _get_sysv_services(). It doesn't
    # matter that it's not part of the _LIST_UNIT_FILES output, we just
    # want to ensure that 'foo' isn't identified as a disabled initscript
    # even though below we are mocking it to show as not enabled (since
    # only 'baz' will be considered an enabled sysv service).
    listdir_mock = MagicMock(return_value=["foo", "bar", "baz", "README"])
    sd_mock = MagicMock(
        return_value={x.replace(".service", "") for x in systemctl_status}
    )
    access_mock = MagicMock(
        side_effect=lambda x, y: x != os.path.join(systemd.INITSCRIPT_PATH, "README")
    )
    sysv_enabled_mock = MagicMock(side_effect=lambda x, _: x == "baz")

    with patch.dict(systemd.__salt__, {"cmd.run": cmd_mock}):
        with patch.object(os, "listdir", listdir_mock):
            with patch.object(systemd, "_get_systemd_services", sd_mock):
                with patch.object(os, "access", side_effect=access_mock):
                    with patch.object(systemd, "_sysv_enabled", sysv_enabled_mock):
                        assert systemd.get_static() == [
                            "service12",
                            "service3",
                            "service6",
                            "service9",
                            "timer12.timer",
                            "timer3.timer",
                            "timer6.timer",
                            "timer9.timer",
                        ]


def test_get_all():
    """
    Test to return a list of all available services
    """
    listdir_mock = MagicMock(
        side_effect=[
            ["foo.service", "multi-user.target.wants", "mytimer.timer"],
            [],
            ["foo.service", "multi-user.target.wants", "bar.service"],
            ["mysql", "nginx", "README"],
        ]
    )
    access_mock = MagicMock(
        side_effect=lambda x, y: x != os.path.join(systemd.INITSCRIPT_PATH, "README")
    )
    with patch.object(os, "listdir", listdir_mock):
        with patch.object(os, "access", side_effect=access_mock):
            assert systemd.get_all() == [
                "bar",
                "foo",
                "mysql",
                "mytimer.timer",
                "nginx",
            ]


def test_available(systemctl_status, systemctl_status_gte_231):
    """
    Test to check that the given service is available
    """
    mock = MagicMock(side_effect=lambda x: systemctl_status[x])

    # systemd < 231
    with patch.dict(systemd.__context__, {"salt.utils.systemd.version": 230}):
        with patch.object(systemd, "_systemctl_status", mock), patch.object(
            systemd, "offline", MagicMock(return_value=False)
        ):
            assert systemd.available("sshd.service") is True
            assert systemd.available("foo.service") is False

    # systemd >= 231
    with patch.dict(systemd.__context__, {"salt.utils.systemd.version": 231}):
        with patch.dict(systemctl_status, systemctl_status_gte_231):
            with patch.object(systemd, "_systemctl_status", mock), patch.object(
                systemd, "offline", MagicMock(return_value=False)
            ):
                assert systemd.available("sshd.service") is True
                assert systemd.available("bar.service") is False

    # systemd < 231 with retcode/output changes backported (e.g. RHEL 7.3)
    with patch.dict(systemd.__context__, {"salt.utils.systemd.version": 219}):
        with patch.dict(systemctl_status, systemctl_status_gte_231):
            with patch.object(systemd, "_systemctl_status", mock), patch.object(
                systemd, "offline", MagicMock(return_value=False)
            ):
                assert systemd.available("sshd.service") is True
                assert systemd.available("bar.service") is False


def test_missing(systemctl_status, systemctl_status_gte_231):
    """
    Test to the inverse of service.available.
    """
    mock = MagicMock(side_effect=lambda x: systemctl_status[x])

    # systemd < 231
    with patch.dict(systemd.__context__, {"salt.utils.systemd.version": 230}):
        with patch.object(systemd, "_systemctl_status", mock), patch.object(
            systemd, "offline", MagicMock(return_value=False)
        ):
            assert systemd.missing("sshd.service") is False
            assert systemd.missing("foo.service") is True

    # systemd >= 231
    with patch.dict(systemd.__context__, {"salt.utils.systemd.version": 231}):
        with patch.dict(systemctl_status, systemctl_status_gte_231):
            with patch.object(systemd, "_systemctl_status", mock), patch.object(
                systemd, "offline", MagicMock(return_value=False)
            ):
                assert systemd.missing("sshd.service") is False
                assert systemd.missing("bar.service") is True

    # systemd < 231 with retcode/output changes backported (e.g. RHEL 7.3)
    with patch.dict(systemd.__context__, {"salt.utils.systemd.version": 219}):
        with patch.dict(systemctl_status, systemctl_status_gte_231):
            with patch.object(systemd, "_systemctl_status", mock), patch.object(
                systemd, "offline", MagicMock(return_value=False)
            ):
                assert systemd.missing("sshd.service") is False
                assert systemd.missing("bar.service") is True


def test_show():
    """
    Test to show properties of one or more units/jobs or the manager
    """
    show_output = "a=b\nc=d\ne={ f=g ; h=i }\nWants=foo.service bar.service\n"
    mock = MagicMock(return_value=show_output)
    with patch.dict(systemd.__salt__, {"cmd.run": mock}):
        assert systemd.show("sshd") == {
            "a": "b",
            "c": "d",
            "e": {"f": "g", "h": "i"},
            "Wants": ["foo.service", "bar.service"],
        }


def test_execs():
    """
    Test to return a list of all files specified as ``ExecStart`` for all
    services
    """
    mock = MagicMock(return_value=["a", "b"])
    with patch.object(systemd, "get_all", mock):
        mock = MagicMock(return_value={"ExecStart": {"path": "c"}})
        with patch.object(systemd, "show", mock):
            assert systemd.execs() == {"a": "c", "b": "c"}


@pytest.fixture()
def unit_name():
    return "foo"


@pytest.fixture()
def mock_none():
    return MagicMock(return_value=None)


@pytest.fixture()
def mock_success():
    return MagicMock(return_value=0)


@pytest.fixture()
def mock_failure():
    return MagicMock(return_value=1)


@pytest.fixture()
def mock_true():
    return MagicMock(return_value=True)


@pytest.fixture()
def mock_false():
    return MagicMock(return_value=False)


@pytest.fixture()
def mock_empty_list():
    return MagicMock(return_value=[])


@pytest.fixture()
def mock_run_all_success():
    return MagicMock(
        return_value={"retcode": 0, "stdout": "", "stderr": "", "pid": 12345}
    )


@pytest.fixture()
def mock_run_all_failure():
    return MagicMock(
        return_value={"retcode": 1, "stdout": "", "stderr": "", "pid": 12345}
    )


@pytest.mark.parametrize(
    "action,no_block",
    [
        ["start", False],
        ["start", True],
        ["stop", False],
        ["stop", True],
        ["restart", False],
        ["restart", True],
        ["reload_", False],
        ["reload_", True],
        ["force_reload", False],
        ["force_reload", True],
        ["enable", False],
        ["enable", True],
        ["disable", False],
        ["disable", True],
    ],
)
def test_change_state(
    unit_name,
    mock_none,
    mock_empty_list,
    mock_true,
    mock_false,
    mock_run_all_success,
    mock_run_all_failure,
    action,
    no_block,
):
    """
    Common code for start/stop/restart/reload/force_reload tests
    """
    # We want the traceback if the function name can't be found in the
    # systemd execution module.
    func = getattr(systemd, action)
    # Remove trailing _ in "reload_"
    action = action.rstrip("_").replace("_", "-")
    systemctl_command = ["/bin/systemctl"]
    if no_block:
        systemctl_command.append("--no-block")
    systemctl_command.extend([action, unit_name + ".service"])
    scope_prefix = ["/bin/systemd-run", "--scope"]

    assert_kwargs = {"python_shell": False}
    if action in ("enable", "disable"):
        assert_kwargs["ignore_retcode"] = True

    with patch("salt.utils.path.which", lambda x: "/bin/" + x):
        with patch.object(systemd, "_check_for_unit_changes", mock_none):
            with patch.object(systemd, "_unit_file_changed", mock_none):
                with patch.object(systemd, "_check_unmask", mock_none):
                    with patch.object(systemd, "_get_sysv_services", mock_empty_list):

                        # Has scopes available
                        with patch.object(salt.utils.systemd, "has_scope", mock_true):

                            # Scope enabled, successful
                            with patch.dict(
                                systemd.__salt__,
                                {
                                    "config.get": mock_true,
                                    "cmd.run_all": mock_run_all_success,
                                },
                            ):
                                ret = func(unit_name, no_block=no_block)
                                assert ret is True
                                mock_run_all_success.assert_called_with(
                                    scope_prefix + systemctl_command, **assert_kwargs
                                )

                            # Scope enabled, failed
                            with patch.dict(
                                systemd.__salt__,
                                {
                                    "config.get": mock_true,
                                    "cmd.run_all": mock_run_all_failure,
                                },
                            ):
                                if action in ("stop", "disable"):
                                    ret = func(unit_name, no_block=no_block)
                                    assert ret is False
                                else:
                                    with pytest.raises(CommandExecutionError):
                                        func(unit_name, no_block=no_block)
                                mock_run_all_failure.assert_called_with(
                                    scope_prefix + systemctl_command, **assert_kwargs
                                )

                            # Scope disabled, successful
                            with patch.dict(
                                systemd.__salt__,
                                {
                                    "config.get": mock_false,
                                    "cmd.run_all": mock_run_all_success,
                                },
                            ):
                                ret = func(unit_name, no_block=no_block)
                                assert ret is True
                                mock_run_all_success.assert_called_with(
                                    systemctl_command, **assert_kwargs
                                )

                            # Scope disabled, failed
                            with patch.dict(
                                systemd.__salt__,
                                {
                                    "config.get": mock_false,
                                    "cmd.run_all": mock_run_all_failure,
                                },
                            ):
                                if action in ("stop", "disable"):
                                    ret = func(unit_name, no_block=no_block)
                                    assert ret is False
                                else:
                                    with pytest.raises(CommandExecutionError):
                                        func(unit_name, no_block=no_block)
                                mock_run_all_failure.assert_called_with(
                                    systemctl_command, **assert_kwargs
                                )

                        # Does not have scopes available
                        with patch.object(salt.utils.systemd, "has_scope", mock_false):

                            # The results should be the same irrespective of
                            # whether or not scope is enabled, since scope is not
                            # available, so we repeat the below tests with it both
                            # enabled and disabled.
                            for scope_mock in (mock_true, mock_false):

                                # Successful
                                with patch.dict(
                                    systemd.__salt__,
                                    {
                                        "config.get": scope_mock,
                                        "cmd.run_all": mock_run_all_success,
                                    },
                                ):
                                    ret = func(unit_name, no_block=no_block)
                                    assert ret is True
                                    mock_run_all_success.assert_called_with(
                                        systemctl_command, **assert_kwargs
                                    )

                                # Failed
                                with patch.dict(
                                    systemd.__salt__,
                                    {
                                        "config.get": scope_mock,
                                        "cmd.run_all": mock_run_all_failure,
                                    },
                                ):
                                    if action in ("stop", "disable"):
                                        ret = func(unit_name, no_block=no_block)
                                        assert ret is False
                                    else:
                                        with pytest.raises(CommandExecutionError):
                                            func(unit_name, no_block=no_block)
                                    mock_run_all_failure.assert_called_with(
                                        systemctl_command, **assert_kwargs
                                    )


@pytest.mark.parametrize(
    "action,runtime",
    [
        ["mask", False],
        ["mask", True],
        ["unmask_", False],
        ["unmask_", True],
    ],
)
def test_mask_unmask(
    unit_name,
    mock_none,
    mock_true,
    mock_false,
    mock_run_all_success,
    mock_run_all_failure,
    action,
    runtime,
):
    """
    Common code for mask/unmask tests
    """
    # We want the traceback if the function name can't be found in the
    # systemd execution module, so don't provide a fallback value for the
    # call to getattr() here.
    func = getattr(systemd, action)
    # Remove trailing _ in "unmask_"
    action = action.rstrip("_").replace("_", "-")
    systemctl_command = ["/bin/systemctl", action]
    if runtime:
        systemctl_command.append("--runtime")
    systemctl_command.append(unit_name + ".service")
    scope_prefix = ["/bin/systemd-run", "--scope"]

    args = [unit_name, runtime]

    masked_mock = mock_true if action == "unmask" else mock_false

    with patch("salt.utils.path.which", lambda x: "/bin/" + x):
        with patch.object(systemd, "_check_for_unit_changes", mock_none):
            if action == "unmask":
                mock_not_run = MagicMock(
                    return_value={
                        "retcode": 0,
                        "stdout": "",
                        "stderr": "",
                        "pid": 12345,
                    }
                )
                with patch.dict(systemd.__salt__, {"cmd.run_all": mock_not_run}):
                    with patch.object(systemd, "masked", mock_false):
                        # Test not masked (should take no action and return True)
                        assert systemd.unmask_(unit_name) is True
                        # Also should not have called cmd.run_all
                        assert mock_not_run.call_count == 0

            with patch.object(systemd, "masked", masked_mock):

                # Has scopes available
                with patch.object(salt.utils.systemd, "has_scope", mock_true):

                    # Scope enabled, successful
                    with patch.dict(
                        systemd.__salt__,
                        {
                            "config.get": mock_true,
                            "cmd.run_all": mock_run_all_success,
                        },
                    ):
                        ret = func(*args)
                        assert ret is True
                        mock_run_all_success.assert_called_with(
                            scope_prefix + systemctl_command,
                            python_shell=False,
                            redirect_stderr=True,
                        )

                    # Scope enabled, failed
                    with patch.dict(
                        systemd.__salt__,
                        {
                            "config.get": mock_true,
                            "cmd.run_all": mock_run_all_failure,
                        },
                    ):
                        with pytest.raises(CommandExecutionError):
                            func(*args)
                        mock_run_all_failure.assert_called_with(
                            scope_prefix + systemctl_command,
                            python_shell=False,
                            redirect_stderr=True,
                        )

                    # Scope disabled, successful
                    with patch.dict(
                        systemd.__salt__,
                        {
                            "config.get": mock_false,
                            "cmd.run_all": mock_run_all_success,
                        },
                    ):
                        ret = func(*args)
                        assert ret is True
                        mock_run_all_success.assert_called_with(
                            systemctl_command,
                            python_shell=False,
                            redirect_stderr=True,
                        )

                    # Scope disabled, failed
                    with patch.dict(
                        systemd.__salt__,
                        {
                            "config.get": mock_false,
                            "cmd.run_all": mock_run_all_failure,
                        },
                    ):
                        with pytest.raises(CommandExecutionError):
                            func(*args)
                        mock_run_all_failure.assert_called_with(
                            systemctl_command,
                            python_shell=False,
                            redirect_stderr=True,
                        )

                # Does not have scopes available
                with patch.object(salt.utils.systemd, "has_scope", mock_false):

                    # The results should be the same irrespective of
                    # whether or not scope is enabled, since scope is not
                    # available, so we repeat the below tests with it both
                    # enabled and disabled.
                    for scope_mock in (mock_true, mock_false):

                        # Successful
                        with patch.dict(
                            systemd.__salt__,
                            {
                                "config.get": scope_mock,
                                "cmd.run_all": mock_run_all_success,
                            },
                        ):
                            ret = func(*args)
                            assert ret is True
                            mock_run_all_success.assert_called_with(
                                systemctl_command,
                                python_shell=False,
                                redirect_stderr=True,
                            )

                        # Failed
                        with patch.dict(
                            systemd.__salt__,
                            {
                                "config.get": scope_mock,
                                "cmd.run_all": mock_run_all_failure,
                            },
                        ):
                            with pytest.raises(CommandExecutionError):
                                func(*args)
                            mock_run_all_failure.assert_called_with(
                                systemctl_command,
                                python_shell=False,
                                redirect_stderr=True,
                            )


def test_firstboot():
    """
    Test service.firstboot without parameters
    """
    result = {"retcode": 0, "stdout": "stdout"}
    salt_mock = {
        "cmd.run_all": MagicMock(return_value=result),
    }
    with patch("salt.utils.path.which", lambda x: "/bin/" + x):
        with patch.dict(systemd.__salt__, salt_mock):
            assert systemd.firstboot()
            salt_mock["cmd.run_all"].assert_called_with(["/bin/systemd-firstboot"])


def test_firstboot_params():
    """
    Test service.firstboot with parameters
    """
    result = {"retcode": 0, "stdout": "stdout"}
    salt_mock = {
        "cmd.run_all": MagicMock(return_value=result),
    }
    with patch("salt.utils.path.which", lambda x: "/bin/" + x):
        with patch.dict(systemd.__salt__, salt_mock):
            assert systemd.firstboot(
                locale="en_US.UTF-8",
                locale_message="en_US.UTF-8",
                keymap="jp",
                timezone="Europe/Berlin",
                hostname="node-001",
                machine_id="1234567890abcdef",
                root="/mnt",
            )
            salt_mock["cmd.run_all"].assert_called_with(
                [
                    "/bin/systemd-firstboot",
                    "--locale",
                    "en_US.UTF-8",
                    "--locale-message",
                    "en_US.UTF-8",
                    "--keymap",
                    "jp",
                    "--timezone",
                    "Europe/Berlin",
                    "--hostname",
                    "node-001",
                    "--machine-ID",
                    "1234567890abcdef",
                    "--root",
                    "/mnt",
                ]
            )


def test_firstboot_error():
    """
    Test service.firstboot error
    """
    result = {"retcode": 1, "stderr": "error"}
    salt_mock = {
        "cmd.run_all": MagicMock(return_value=result),
    }
    with patch.dict(systemd.__salt__, salt_mock):
        with pytest.raises(CommandExecutionError):
            assert systemd.firstboot()



"""
Zeromq transport classes
"""

import asyncio
import asyncio.exceptions
import errno
import hashlib
import logging
import multiprocessing
import os
import signal
import sys
import threading
from random import randint

import tornado
import tornado.concurrent
import tornado.gen
import tornado.ioloop
import tornado.locks
import zmq.asyncio
import zmq.error
import zmq.eventloop.future
import zmq.eventloop.zmqstream

import salt.payload
import salt.transport.base
import salt.utils.files
import salt.utils.process
import salt.utils.stringutils
import salt.utils.zeromq
from salt._compat import ipaddress
from salt.exceptions import SaltException, SaltReqTimeoutError
from salt.utils.zeromq import LIBZMQ_VERSION_INFO, ZMQ_VERSION_INFO, zmq

try:
    import zmq.utils.monitor

    HAS_ZMQ_MONITOR = True
except ImportError:
    HAS_ZMQ_MONITOR = False


log = logging.getLogger(__name__)


def _get_master_uri(master_ip, master_port, source_ip=None, source_port=None):
    """
    Return the ZeroMQ URI to connect the Minion to the Master.
    It supports different source IP / port, given the ZeroMQ syntax:
    // Connecting using a IP address and bind to an IP address
    rc = zmq_connect(socket, "tcp://192.168.1.17:5555;192.168.1.1:5555"); assert (rc == 0);
    Source: http://api.zeromq.org/4-1:zmq-tcp
    """
    from salt.utils.network import ip_bracket

    master_uri = "tcp://{master_ip}:{master_port}".format(
        master_ip=ip_bracket(master_ip), master_port=master_port
    )

    if source_ip or source_port:
        if LIBZMQ_VERSION_INFO >= (4, 1, 6) and ZMQ_VERSION_INFO >= (16, 0, 1):
            # The source:port syntax for ZeroMQ has been added in libzmq 4.1.6
            # which is included in the pyzmq wheels starting with 16.0.1.
            if source_ip and source_port:
                master_uri = (
                    "tcp://{source_ip}:{source_port};{master_ip}:{master_port}".format(
                        source_ip=ip_bracket(source_ip),
                        source_port=source_port,
                        master_ip=ip_bracket(master_ip),
                        master_port=master_port,
                    )
                )
            elif source_ip and not source_port:
                master_uri = "tcp://{source_ip}:0;{master_ip}:{master_port}".format(
                    source_ip=ip_bracket(source_ip),
                    master_ip=ip_bracket(master_ip),
                    master_port=master_port,
                )
            elif source_port and not source_ip:
                ip_any = (
                    "0.0.0.0"
                    if ipaddress.ip_address(master_ip).version == 4
                    else ip_bracket("::")
                )
                master_uri = (
                    "tcp://{ip_any}:{source_port};{master_ip}:{master_port}".format(
                        ip_any=ip_any,
                        source_port=source_port,
                        master_ip=ip_bracket(master_ip),
                        master_port=master_port,
                    )
                )
        else:
            log.warning(
                "Unable to connect to the Master using a specific source IP / port"
            )
            log.warning("Consider upgrading to pyzmq >= 16.0.1 and libzmq >= 4.1.6")
            log.warning(
                "Specific source IP / port for connecting to master returner port:"
                " configuraion ignored"
            )

    return master_uri


class PublishClient(salt.transport.base.PublishClient):
    """
    A transport channel backed by ZeroMQ for a Salt Publisher to use to
    publish commands to connected minions
    """

    ttype = "zeromq"

    async_methods = [
        "connect",
        "connect_uri",
        "recv",
        # "close",
    ]
    close_methods = [
        "close",
    ]

    def _legacy_setup(
        self,
        _id,
        role,
        zmq_filtering=False,
        tcp_keepalive=True,
        tcp_keepalive_idle=300,
        tcp_keepalive_cnt=-1,
        tcp_keepalive_intvl=-1,
        recon_default=1000,
        recon_max=10000,
        recon_randomize=True,
        ipv6=None,
        master_ip="127.0.0.1",
        zmq_monitor=False,
        **extras,
    ):
        self.hexid = hashlib.sha1(salt.utils.stringutils.to_bytes(_id)).hexdigest()
        self._closing = False
        self.context = zmq.asyncio.Context()
        self._socket = self.context.socket(zmq.SUB)
        self._socket.setsockopt(zmq.LINGER, -1)
        if zmq_filtering:
            # TODO: constants file for "broadcast"
            self._socket.setsockopt(zmq.SUBSCRIBE, b"broadcast")
            if role == "syndic":
                self._socket.setsockopt(zmq.SUBSCRIBE, b"syndic")
            else:
                self._socket.setsockopt(
                    zmq.SUBSCRIBE, salt.utils.stringutils.to_bytes(self.hexid)
                )
        else:
            self._socket.setsockopt(zmq.SUBSCRIBE, b"")

        if _id:
            self._socket.setsockopt(zmq.IDENTITY, salt.utils.stringutils.to_bytes(_id))

        # TODO: cleanup all the socket opts stuff
        if hasattr(zmq, "TCP_KEEPALIVE"):
            self._socket.setsockopt(zmq.TCP_KEEPALIVE, tcp_keepalive)
            self._socket.setsockopt(zmq.TCP_KEEPALIVE_IDLE, tcp_keepalive_idle)
            self._socket.setsockopt(zmq.TCP_KEEPALIVE_CNT, tcp_keepalive_cnt)
            self._socket.setsockopt(zmq.TCP_KEEPALIVE_INTVL, tcp_keepalive_intvl)

        if recon_randomize:
            recon_delay = randint(
                recon_default,
                recon_default + recon_max,
            )

            log.debug(
                "Generated random reconnect delay between '%sms' and '%sms' (%s)",
                recon_delay,
                recon_delay + recon_max,
                recon_delay,
            )

            log.debug("Setting zmq_reconnect_ivl to '%sms'", recon_delay)
            self._socket.setsockopt(zmq.RECONNECT_IVL, recon_delay)

            if hasattr(zmq, "RECONNECT_IVL_MAX"):
                log.debug(
                    "Setting zmq_reconnect_ivl_max to '%sms'",
                    recon_delay + recon_max,
                )

                self._socket.setsockopt(zmq.RECONNECT_IVL_MAX, recon_max)

        if (ipv6 is True or ":" in master_ip) and hasattr(zmq, "IPV4ONLY"):
            # IPv6 sockets work for both IPv6 and IPv4 addresses
            self._socket.setsockopt(zmq.IPV4ONLY, 0)

        self.poller = zmq.Poller()
        self.poller.register(self._socket, zmq.POLLIN)

        if HAS_ZMQ_MONITOR and zmq_monitor:
            self._monitor = ZeroMQSocketMonitor(self._socket)
            self._monitor.start_io_loop(self.io_loop)

    def __init__(self, opts, io_loop, **kwargs):
        super().__init__(opts, io_loop, **kwargs)
        self.opts = opts
        self.io_loop = io_loop
        self._legacy_setup(
            _id=opts.get("id", ""),
            role=opts.get("__role", ""),
            **opts,
        )
        self.connect_called = False
        self.callbacks = {}

        self.host = kwargs.get("host", None)
        self.port = kwargs.get("port", None)
        self.path = kwargs.get("path", None)
        self.source_ip = self.opts.get("source_ip")
        self.source_port = self.opts.get("source_publish_port")
        if self.host is None and self.port is None:
            if self.path is None:
                raise Exception("A host and port or a path must be provided")
        elif self.host and self.port:
            if self.path:
                raise Exception("A host and port or a path must be provided, not both")
        self.on_recv_task = None

    def close(self):
        if self._closing is True:
            return
        self._closing = True
        if hasattr(self, "_monitor") and self._monitor is not None:
            self._monitor.stop()
            self._monitor = None
        if hasattr(self, "_stream"):
            self._stream.close(0)
        elif hasattr(self, "_socket"):
            self._socket.close(0)
        if hasattr(self, "context") and self.context.closed is False:
            self.context.term()
        callbacks = self.callbacks
        self.callbacks = {}
        for callback, (running, task) in callbacks.items():
            running.clear()
        return

    # pylint: enable=W1701
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    # TODO: this is the time to see if we are connected, maybe use the req channel to guess?
    async def connect(
        self, port=None, connect_callback=None, disconnect_callback=None, timeout=None
    ):
        self._connect_called = True
        if port is not None:
            self.port = port
        if self.path:
            pub_uri = f"ipc://{self.path}"
            log.debug("Connecting the publisher client to: %s", pub_uri)
            self._socket.connect(pub_uri)
        else:
            # host = self.opts["master_ip"],
            if port is not None:
                self.port = port
            master_pub_uri = _get_master_uri(
                self.host, self.port, self.source_ip, self.source_port
            )
            log.debug(
                "Connecting the Minion to the Master publish port, using the URI: %s",
                master_pub_uri,
            )
            self._socket.connect(master_pub_uri)
        if connect_callback:
            await connect_callback(True)

    async def connect_uri(self, uri, connect_callback=None, disconnect_callback=None):
        self._connect_called = True
        log.debug("Connecting the publisher client to: %s", uri)
        # log.debug("%r connecting to %s", self, self.master_pub)
        self.uri = uri
        self._socket.connect(uri)
        if connect_callback:
            await connect_callback(True)

    def _decode_messages(self, messages):
        """
        Take the zmq messages, decrypt/decode them into a payload

        :param list messages: A list of messages to be decoded
        """
        if isinstance(messages, list):
            messages_len = len(messages)
            # if it was one message, then its old style
            if messages_len == 1:
                payload = salt.payload.loads(messages[0])
            # 2 includes a header which says who should do it
            elif messages_len == 2:
                message_target = salt.utils.stringutils.to_str(messages[0])
                if (
                    self.opts.get("__role") != "syndic"
                    and message_target not in ("broadcast", self.hexid)
                ) or (
                    self.opts.get("__role") == "syndic"
                    and message_target not in ("broadcast", "syndic")
                ):
                    log.debug(
                        "Publish received for not this minion: %s", message_target
                    )
                    return None
                payload = salt.payload.loads(messages[1])
            else:
                raise Exception(
                    "Invalid number of messages ({}) in zeromq pubmessage from master".format(
                        len(messages_len)
                    )
                )
        else:
            payload = salt.payload.loads(messages)
        # Yield control back to the caller. When the payload has been decoded, assign
        # the decoded payload to 'ret' and resume operation
        return payload

    async def recv(self, timeout=None):
        if timeout == 0:
            events = self.poller.poll(timeout=timeout)
            if events:
                return await self._socket.recv()
        elif timeout:
            try:
                return await asyncio.wait_for(self._socket.recv(), timeout=timeout)
            except asyncio.exceptions.TimeoutError:
                log.trace("PublishClient recieve timedout: %d", timeout)
        else:
            return await self._socket.recv()

    async def send(self, msg):
        return
        # raise Exception("Send not supported")
        # await self._socket.send(msg)

    # async def on_recv_handler(self, callback):
    #    while not self._socket:
    #        # Retry quickly, we may want to increase this if it's hogging cpu.
    #        await asyncio.sleep(0.003)
    #    while True:
    #        msg = await self.recv()
    #        if msg:
    #            await callback(msg)

    # def on_recv(self, callback):
    #    """
    #    Register a callback for received messages (that we didn't initiate)
    #    """
    #    if self.on_recv_task:
    #        # XXX: We are not awaiting this canceled task. This still needs to
    #        # be addressed.
    #        self.on_recv_task.cancel()
    #    if callback is None:
    #        self.on_recv_task = None
    #    else:
    #        self.on_recv_task = asyncio.create_task(self.on_recv_handler(callback))

    def on_recv(self, callback):
        """
        Register a callback for received messages (that we didn't initiate)

        :param func callback: A function which should be called when data is received
        """
        if callback is None:
            callbacks = self.callbacks
            self.callbacks = {}
            for callback, (running, task) in callbacks.items():
                running.clear()
            return

        running = asyncio.Event()
        running.set()

        async def consume(running):
            try:
                while running.is_set():
                    try:
                        msg = await self.recv(timeout=None)
                    except zmq.error.ZMQError as exc:
                        # We've disconnected just die
                        break
                    if msg:
                        try:
                            await callback(msg)
                        except Exception:  # pylint: disable=broad-except
                            log.error("Exception while running callback", exc_info=True)
                    # log.debug("Callback done %r", callback)
            except Exception as exc:  # pylint: disable=broad-except
                log.error(
                    "Exception while consuming%s %s", self.uri, exc, exc_info=True
                )

        task = self.io_loop.spawn_callback(consume, running)
        self.callbacks[callback] = running, task


class RequestServer(salt.transport.base.DaemonizedRequestServer):
    def __init__(self, opts):  # pylint: disable=W0231
        self.opts = opts
        self._closing = False
        self._monitor = None
        self._w_monitor = None
        self.tasks = set()
        self._event = asyncio.Event()

    def zmq_device(self):
        """
        Multiprocessing target for the zmq queue device
        """
        self.__setup_signals()
        context = zmq.Context(self.opts["worker_threads"])
        # Prepare the zeromq sockets
        self.uri = "tcp://{interface}:{ret_port}".format(**self.opts)
        self.clients = context.socket(zmq.ROUTER)
        self.clients.setsockopt(zmq.LINGER, -1)
        if self.opts["ipv6"] is True and hasattr(zmq, "IPV4ONLY"):
            # IPv6 sockets work for both IPv6 and IPv4 addresses
            self.clients.setsockopt(zmq.IPV4ONLY, 0)
        self.clients.setsockopt(zmq.BACKLOG, self.opts.get("zmq_backlog", 1000))
        self._start_zmq_monitor()
        self.workers = context.socket(zmq.DEALER)
        self.workers.setsockopt(zmq.LINGER, -1)

        if self.opts["mworker_queue_niceness"] and not salt.utils.platform.is_windows():
            log.info(
                "setting mworker_queue niceness to %d",
                self.opts["mworker_queue_niceness"],
            )
            os.nice(self.opts["mworker_queue_niceness"])

        if self.opts.get("ipc_mode", "") == "tcp":
            self.w_uri = "tcp://127.0.0.1:{}".format(
                self.opts.get("tcp_master_workers", 4515)
            )
        else:
            self.w_uri = "ipc://{}".format(
                os.path.join(self.opts["sock_dir"], "workers.ipc")
            )

        log.info("Setting up the master communication server")
        log.info("ReqServer clients %s", self.uri)
        self.clients.bind(self.uri)
        log.info("ReqServer workers %s", self.w_uri)
        self.workers.bind(self.w_uri)
        if self.opts.get("ipc_mode", "") != "tcp":
            os.chmod(os.path.join(self.opts["sock_dir"], "workers.ipc"), 0o600)

        while True:
            if self.clients.closed or self.workers.closed:
                break
            try:
                zmq.device(zmq.QUEUE, self.clients, self.workers)
            except zmq.ZMQError as exc:
                if exc.errno == errno.EINTR:
                    continue
                raise
            except (KeyboardInterrupt, SystemExit):
                break
        context.term()

    def close(self):
        """
        Cleanly shutdown the router socket
        """
        if self._closing:
            return
        log.info("MWorkerQueue under PID %s is closing", os.getpid())
        self._closing = True
        self._event.set()
        if getattr(self, "_monitor", None) is not None:
            self._monitor.stop()
            self._monitor = None
        if getattr(self, "_w_monitor", None) is not None:
            self._w_monitor.stop()
            self._w_monitor = None
        if hasattr(self, "clients") and self.clients.closed is False:
            self.clients.close()
        if hasattr(self, "workers") and self.workers.closed is False:
            self.workers.close()
        if hasattr(self, "stream"):
            self.stream.close()
        if hasattr(self, "_socket") and self._socket.closed is False:
            self._socket.close()
        if hasattr(self, "context") and self.context.closed is False:
            self.context.term()
        for task in list(self.tasks):
            try:
                task.cancel()
            except RuntimeError:
                log.error("IOLoop closed when trying to cancel task")

    def pre_fork(self, process_manager):
        """
        Pre-fork we need to create the zmq router device

        :param func process_manager: An instance of salt.utils.process.ProcessManager
        """
        process_manager.add_process(self.zmq_device, name="MWorkerQueue")

    def _start_zmq_monitor(self):
        """
        Starts ZMQ monitor for debugging purposes.
        :return:
        """
        # Socket monitor shall be used the only for debug
        # purposes so using threading doesn't look too bad here

        if HAS_ZMQ_MONITOR and self.opts["zmq_monitor"]:
            log.debug("Starting ZMQ monitor")
            self._w_monitor = ZeroMQSocketMonitor(self._socket)
            threading.Thread(target=self._w_monitor.start_poll).start()
            log.debug("ZMQ monitor has been started started")

    def post_fork(self, message_handler, io_loop):
        """
        After forking we need to create all of the local sockets to listen to the
        router

        :param func message_handler: A function to called to handle incoming payloads as
                                     they are picked up off the wire
        :param IOLoop io_loop: An instance of a Tornado IOLoop, to handle event scheduling
        """
        # context = zmq.Context(1)
        self.context = zmq.asyncio.Context(1)
        self._socket = self.context.socket(zmq.REP)
        # Linger -1 means we'll never discard messages.
        self._socket.setsockopt(zmq.LINGER, -1)
        self._start_zmq_monitor()

        if self.opts.get("ipc_mode", "") == "tcp":
            self.w_uri = "tcp://127.0.0.1:{}".format(
                self.opts.get("tcp_master_workers", 4515)
            )
        else:
            self.w_uri = "ipc://{}".format(
                os.path.join(self.opts["sock_dir"], "workers.ipc")
            )
        log.info("Worker binding to socket %s", self.w_uri)
        self._socket.connect(self.w_uri)
        if self.opts.get("ipc_mode", "") != "tcp" and os.path.isfile(
            os.path.join(self.opts["sock_dir"], "workers.ipc")
        ):
            os.chmod(os.path.join(self.opts["sock_dir"], "workers.ipc"), 0o600)
        self.message_handler = message_handler

        async def callback():
            task = asyncio.create_task(self.request_handler())
            task.add_done_callback(self.tasks.discard)
            self.tasks.add(task)

        io_loop.add_callback(callback)

    async def request_handler(self):
        while not self._event.is_set():
            try:
                request = await asyncio.wait_for(self._socket.recv(), 0.3)
                reply = await self.handle_message(None, request)
                await self._socket.send(self.encode_payload(reply))
            except asyncio.exceptions.TimeoutError:
                continue
            except Exception as exc:  # pylint: disable=broad-except
                log.error("Exception in request handler", exc_info=True)
                break

    async def handle_message(self, stream, payload):
        try:
            payload = self.decode_payload(payload)
        except salt.exceptions.SaltDeserializationError:
            return {"msg": "bad load"}
        return await self.message_handler(payload)

    def encode_payload(self, payload):
        return salt.payload.dumps(payload)

    def __setup_signals(self):
        signal.signal(signal.SIGINT, self._handle_signals)
        signal.signal(signal.SIGTERM, self._handle_signals)

    def _handle_signals(self, signum, sigframe):
        msg = f"{self.__class__.__name__} received a "
        if signum == signal.SIGINT:
            msg += "SIGINT"
        elif signum == signal.SIGTERM:
            msg += "SIGTERM"
        msg += ". Exiting"
        log.debug(msg)
        self.close()
        sys.exit(salt.defaults.exitcodes.EX_OK)

    def decode_payload(self, payload):
        payload = salt.payload.loads(payload)
        return payload


def _set_tcp_keepalive(zmq_socket, opts):
    """
    Ensure that TCP keepalives are set as specified in "opts".

    Warning: Failure to set TCP keepalives on the salt-master can result in
    not detecting the loss of a minion when the connection is lost or when
    its host has been terminated without first closing the socket.
    Salt's Presence System depends on this connection status to know if a minion
    is "present".

    Warning: Failure to set TCP keepalives on minions can result in frequent or
    unexpected disconnects!
    """
    if hasattr(zmq, "TCP_KEEPALIVE") and opts:
        if "tcp_keepalive" in opts:
            zmq_socket.setsockopt(zmq.TCP_KEEPALIVE, opts["tcp_keepalive"])
        if "tcp_keepalive_idle" in opts:
            zmq_socket.setsockopt(zmq.TCP_KEEPALIVE_IDLE, opts["tcp_keepalive_idle"])
        if "tcp_keepalive_cnt" in opts:
            zmq_socket.setsockopt(zmq.TCP_KEEPALIVE_CNT, opts["tcp_keepalive_cnt"])
        if "tcp_keepalive_intvl" in opts:
            zmq_socket.setsockopt(zmq.TCP_KEEPALIVE_INTVL, opts["tcp_keepalive_intvl"])


class AsyncReqMessageClient:
    """
    This class wraps the underlying zeromq REQ socket and gives a future-based
    interface to sending and recieving messages. This works around the primary
    limitation of serialized send/recv on the underlying socket by queueing the
    message sends in this class. In the future if we decide to attempt to multiplex
    we can manage a pool of REQ/REP sockets-- but for now we'll just do them in serial
    """

    def __init__(self, opts, addr, linger=0, io_loop=None):
        """
        Create an asynchronous message client

        :param dict opts: The salt opts dictionary
        :param str addr: The interface IP address to bind to
        :param int linger: The number of seconds to linger on a ZMQ socket. See
                           http://api.zeromq.org/2-1:zmq-setsockopt [ZMQ_LINGER]
        :param IOLoop io_loop: A Tornado IOLoop event scheduler [tornado.ioloop.IOLoop]
        """
        salt.utils.versions.warn_until(
            3009,
            "AsyncReqMessageClient has been deprecated and will be removed.",
        )
        self.opts = opts
        self.addr = addr
        self.linger = linger
        if io_loop is None:
            self.io_loop = tornado.ioloop.IOLoop.current()
        else:
            self.io_loop = io_loop

        self.context = zmq.eventloop.future.Context()

        self.send_queue = []

        self._closing = False
        self.lock = tornado.locks.Lock()

    def connect(self):
        if hasattr(self, "socket") and self.socket:
            return
        # wire up sockets
        self._init_socket()

    def close(self):
        if self._closing:
            return
        else:
            self._closing = True
            if hasattr(self, "socket") and self.socket is not None:
                self.socket.close(0)
                self.socket = None
            if self.context.closed is False:
                self.context.term()

    def _init_socket(self):
        self.socket = self.context.socket(zmq.REQ)

        # socket options
        if hasattr(zmq, "RECONNECT_IVL_MAX"):
            self.socket.setsockopt(zmq.RECONNECT_IVL_MAX, 5000)

        _set_tcp_keepalive(self.socket, self.opts)
        if self.addr.startswith("tcp://["):
            # Hint PF type if bracket enclosed IPv6 address
            if hasattr(zmq, "IPV6"):
                self.socket.setsockopt(zmq.IPV6, 1)
            elif hasattr(zmq, "IPV4ONLY"):
                self.socket.setsockopt(zmq.IPV4ONLY, 0)
        self.socket.setsockopt(zmq.LINGER, self.linger)
        self.socket.connect(self.addr)

    @tornado.gen.coroutine
    def send(self, message, timeout=None, callback=None):
        """
        Return a future which will be completed when the message has a response
        """
        future = tornado.concurrent.Future()

        message = salt.payload.dumps(message)

        if callback is not None:

            def handle_future(future):
                response = future.result()
                self.io_loop.add_callback(callback, response)

            future.add_done_callback(handle_future)

        if self.opts.get("detect_mode") is True:
            timeout = 1

        if timeout is not None:
            send_timeout = self.io_loop.call_later(
                timeout, self._timeout_message, future
            )

        self.io_loop.spawn_callback(self._send_recv, message, future)

        recv = yield future

        raise tornado.gen.Return(recv)

    def _timeout_message(self, future):
        if not future.done():
            future.set_exception(SaltReqTimeoutError("Message timed out"))

    @tornado.gen.coroutine
    def _send_recv(self, message, future):
        try:
            with (yield self.lock.acquire()):
                yield self.socket.send(message)
                try:
                    recv = yield self.socket.recv()
                except zmq.eventloop.future.CancelledError as exc:
                    if not future.done():
                        future.set_exception(exc)
                    return

            if not future.done():
                data = salt.payload.loads(recv)
                future.set_result(data)
        except Exception as exc:  # pylint: disable=broad-except
            if not future.done():
                future.set_exception(exc)


class ZeroMQSocketMonitor:
    __EVENT_MAP = None

    def __init__(self, socket):
        """
        Create ZMQ monitor sockets

        More information:
            http://api.zeromq.org/4-0:zmq-socket-monitor
        """
        self._socket = socket
        self._monitor_socket = self._socket.get_monitor_socket()
        self._monitor_task = None
        self._running = asyncio.Event()

    def start_io_loop(self, io_loop):
        log.trace("Event monitor start!")
        self._running.set()
        io_loop.spawn_callback(self.consume)

    async def consume(self):
        while self._running.is_set():
            try:
                if await self._monitor_socket.poll():
                    msg = await self._monitor_socket.recv_multipart()
                    self.monitor_callback(msg)
                else:
                    await asyncio.sleep(0.3)
            except zmq.error.ZMQError as exc:
                log.error("ZmqMonitor, %s", exc)
                # We've disconnected just die
                break
            except Exception as exc:  # pylint: disable=broad-except
                log.error("ZmqMonitor, %s", exc)
                break

    def start_poll(self):
        log.trace("Event monitor start!")
        try:
            while self._monitor_socket is not None and self._monitor_socket.poll():
                msg = self._monitor_socket.recv_multipart()
                self.monitor_callback(msg)
        except (AttributeError, zmq.error.ContextTerminated):
            # We cannot log here because we'll get an interrupted system call in trying
            # to flush the logging buffer as we terminate
            pass

    @property
    def event_map(self):
        if ZeroMQSocketMonitor.__EVENT_MAP is None:
            event_map = {}
            for name in dir(zmq):
                if name.startswith("EVENT_"):
                    value = getattr(zmq, name)
                    event_map[value] = name
            ZeroMQSocketMonitor.__EVENT_MAP = event_map
        return ZeroMQSocketMonitor.__EVENT_MAP

    def monitor_callback(self, msg):
        evt = zmq.utils.monitor.parse_monitor_message(msg)
        evt["description"] = self.event_map[evt["event"]]
        log.debug("ZeroMQ event: %s", evt)
        if evt["event"] == zmq.EVENT_MONITOR_STOPPED:
            self.stop()

    def stop(self):
        if self._socket is None:
            return
        try:
            self._socket.disable_monitor()
        except zmq.Error:
            pass
        self._socket = None
        self._running.clear()
        self._monitor_socket = None
        log.trace("Event monitor done!")


class PublishServer(salt.transport.base.DaemonizedPublishServer):
    """
    Encapsulate synchronous operations for a publisher channel
    """

    async_methods = [
        "publish",
    ]
    close_methods = [
        "close",
    ]

    def __init__(
        self,
        opts,
        pub_host=None,
        pub_port=None,
        pub_path=None,
        pull_host=None,
        pull_port=None,
        pull_path=None,
        pull_path_perms=0o600,
        pub_path_perms=0o600,
        started=None,
    ):
        self.opts = opts
        self.pub_host = pub_host
        self.pub_port = pub_port
        self.pub_path = pub_path
        if pub_path:
            self.pub_uri = f"ipc://{pub_path}"
        else:
            self.pub_uri = f"tcp://{pub_host}:{pub_port}"
        self.pull_host = pull_host
        self.pull_port = pull_port
        self.pull_path = pull_path
        self.pub_path_perms = pub_path_perms
        self.pull_path_perms = pull_path_perms
        if pull_path:
            self.pull_uri = f"ipc://{pull_path}"
        else:
            self.pull_uri = f"tcp://{pull_host}:{pull_port}"
        self.ctx = None
        self.sock = None
        self.daemon_context = None
        self.daemon_pub_sock = None
        self.daemon_pull_sock = None
        self.daemon_monitor = None
        if started is None:
            self.started = multiprocessing.Event()
        else:
            self.started = started

    def __repr__(self):
        return f"<PublishServer pub_uri={self.pub_uri} pull_uri={self.pull_uri} at {hex(id(self))}>"

    def __setstate__(self, state):
        self.__init__(**state)

    def __getstate__(self):
        return {
            "opts": self.opts,
            "pub_host": self.pub_host,
            "pub_port": self.pub_port,
            "pub_path": self.pub_path,
            "pull_host": self.pull_host,
            "pull_port": self.pull_port,
            "pull_path": self.pull_path,
            "pub_path_perms": self.pub_path_perms,
            "pull_path_perms": self.pull_path_perms,
            "started": self.started,
        }

    def publish_daemon(
        self,
        publish_payload,
        presence_callback=None,
        remove_presence_callback=None,
    ):
        """
        This method represents the Publish Daemon process. It is intended to be
        run in a thread or process as it creates and runs its own ioloop.
        """
        ioloop = tornado.ioloop.IOLoop()
        ioloop.add_callback(self.publisher, publish_payload, ioloop=ioloop)
        try:
            ioloop.start()
        finally:
            self.close()

    def _get_sockets(self, context, ioloop):
        pub_sock = context.socket(zmq.PUB)
        monitor = ZeroMQSocketMonitor(pub_sock)
        monitor.start_io_loop(ioloop)
        _set_tcp_keepalive(pub_sock, self.opts)
        self.dpub_sock = pub_sock  # = zmq.eventloop.zmqstream.ZMQStream(pub_sock)
        # if 2.1 >= zmq < 3.0, we only have one HWM setting
        try:
            pub_sock.setsockopt(zmq.HWM, self.opts.get("pub_hwm", 1000))
        # in zmq >= 3.0, there are separate send and receive HWM settings
        except (AttributeError, zmq.error.ZMQError):
            # Set the High Water Marks. For more information on HWM, see:
            # http://api.zeromq.org/4-1:zmq-setsockopt
            pub_sock.setsockopt(zmq.SNDHWM, self.opts.get("pub_hwm", 1000))
            pub_sock.setsockopt(zmq.RCVHWM, self.opts.get("pub_hwm", 1000))
        if self.opts["ipv6"] is True and hasattr(zmq, "IPV4ONLY"):
            # IPv6 sockets work for both IPv6 and IPv4 addresses
            pub_sock.setsockopt(zmq.IPV4ONLY, 0)

        pub_sock.setsockopt(zmq.BACKLOG, self.opts.get("zmq_backlog", 1000))
        pub_sock.setsockopt(zmq.LINGER, -1)
        # Prepare minion pull socket
        pull_sock = context.socket(zmq.PULL)
        pull_sock.setsockopt(zmq.LINGER, -1)
        # pull_sock = zmq.eventloop.zmqstream.ZMQStream(pull_sock)
        pull_sock.setsockopt(zmq.LINGER, -1)
        salt.utils.zeromq.check_ipc_path_max_len(self.pull_uri)
        # Start the minion command publisher
        # Securely create socket
        with salt.utils.files.set_umask(0o177):
            log.info("Starting the Salt Publisher on %s", self.pub_uri)
            pub_sock.bind(self.pub_uri)
            if self.pub_path:
                os.chmod(  # nosec
                    self.pub_path,
                    self.pub_path_perms,
                )
            log.info("Starting the Salt Puller on %s", self.pull_uri)
            pull_sock.bind(self.pull_uri)
            if self.pull_path:
                os.chmod(  # nosec
                    self.pull_path,
                    self.pull_path_perms,
                )
        return pull_sock, pub_sock, monitor

    async def publisher(self, publish_payload, ioloop=None):
        if ioloop is None:
            ioloop = tornado.ioloop.IOLoop.current()
        self.daemon_context = zmq.asyncio.Context()
        (
            self.daemon_pull_sock,
            self.daemon_pub_sock,
            self.daemon_monitor,
        ) = self._get_sockets(self.daemon_context, ioloop)
        self.started.set()
        while True:
            try:
                package = await self.daemon_pull_sock.recv()
                await publish_payload(package)
            except Exception as exc:  # pylint: disable=broad-except
                log.error(
                    "Exception in publisher %s %s",
                    self.pull_uri,
                    exc,
                    exc_info_on_loglevel=logging.DEBUG,
                )

    async def publish_payload(self, payload, topic_list=None):
        log.trace("Publish payload %r", payload)
        if self.opts["zmq_filtering"]:
            if topic_list:
                for topic in topic_list:
                    log.trace("Sending filtered data over publisher %s", self.pub_uri)
                    # zmq filters are substring match, hash the topic
                    # to avoid collisions
                    htopic = salt.utils.stringutils.to_bytes(
                        hashlib.sha1(salt.utils.stringutils.to_bytes(topic)).hexdigest()
                    )
                    await self.dpub_sock.send_multipart([htopic, payload])
                    log.trace("Filtered data has been sent")
                # Syndic broadcast
                if self.opts.get("order_masters"):
                    log.trace("Sending filtered data to syndic")
                    await self.dpub_sock.send_multipart([b"syndic", payload])
                    log.trace("Filtered data has been sent to syndic")
            # otherwise its a broadcast
            else:
                # TODO: constants file for "broadcast"
                log.trace("Sending broadcasted data over publisher %s", self.pub_uri)
                await self.dpub_sock.send_multipart([b"broadcast", payload])
                log.trace("Broadcasted data has been sent")
        else:
            log.trace("Sending ZMQ-unfiltered data over publisher %s", self.pub_uri)
            await self.dpub_sock.send(payload)
            log.trace("Unfiltered data has been sent")

    def pre_fork(self, process_manager):
        """
        Do anything necessary pre-fork. Since this is on the master side this will
        primarily be used to create IPC channels and create our daemon process to
        do the actual publishing

        :param func process_manager: A ProcessManager, from salt.utils.process.ProcessManager
        """
        process_manager.add_process(
            self.publish_daemon,
            args=(self.publish_payload,),
        )

    def connect(self, timeout=None):
        """
        Create and connect this thread's zmq socket. If a publisher socket
        already exists "pub_close" is called before creating and connecting a
        new socket.
        """
        log.debug("Connecting to pub server: %s", self.pull_uri)
        self.ctx = zmq.asyncio.Context()
        self.sock = self.ctx.socket(zmq.PUSH)
        self.sock.setsockopt(zmq.LINGER, -1)
        self.sock.connect(self.pull_uri)
        return self.sock

    def close(self):
        """
        Disconnect an existing publisher socket and remove it from the local
        thread's cache.
        """
        if self.sock is not None:
            sock = self.sock
            self.sock = None
            sock.close()
        if self.ctx and self.ctx.closed is False:
            ctx = self.ctx
            self.ctx = None
            ctx.term()
        if self.daemon_monitor:
            self.daemon_monitor.stop()
        if self.daemon_pub_sock:
            self.daemon_pub_sock.close()
        if self.daemon_pull_sock:
            self.daemon_pull_sock.close()
        if self.daemon_context:
            self.daemon_context.destroy(1)
            self.daemon_context.term()

    async def publish(
        self, payload, **kwargs
    ):  # pylint: disable=invalid-overridden-method
        """
        Publish "load" to minions. This send the load to the publisher daemon
        process with does the actual sending to minions.

        :param dict load: A load to be sent across the wire to minions
        """
        if not self.sock:
            self.connect()
        await self.sock.send(payload)

    @property
    def topic_support(self):
        return self.opts.get("zmq_filtering", False)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class RequestClient(salt.transport.base.RequestClient):

    ttype = "zeromq"

    def __init__(self, opts, io_loop, linger=0):  # pylint: disable=W0231
        super().__init__(opts, io_loop)
        self.opts = opts
        # XXX Support host, port, path, instead of using get_master_uri
        self.master_uri = self.get_master_uri(opts)
        self.linger = linger
        if io_loop is None:
            self.io_loop = tornado.ioloop.IOLoop.current()
        else:
            self.io_loop = io_loop
        self.context = None
        self.send_queue = []
        # mapping of message -> future
        self.send_future_map = {}
        self._closing = False
        self.socket = None
        self.sending = asyncio.Lock()

    async def connect(self):  # pylint: disable=invalid-overridden-method
        if self.socket is None:
            self._connect_called = True
            self._closing = False
            # wire up sockets
            self._init_socket()

    def _init_socket(self):
        if self.socket is not None:
            self.context = zmq.asyncio.Context()
            self.socket.close()  # pylint: disable=E0203
            del self.socket
        self.context = zmq.asyncio.Context()
        self.socket = self.context.socket(zmq.REQ)
        self.socket.setsockopt(zmq.LINGER, -1)

        # socket options
        if hasattr(zmq, "RECONNECT_IVL_MAX"):
            self.socket.setsockopt(zmq.RECONNECT_IVL_MAX, 5000)

        _set_tcp_keepalive(self.socket, self.opts)
        if self.master_uri.startswith("tcp://["):
            # Hint PF type if bracket enclosed IPv6 address
            if hasattr(zmq, "IPV6"):
                self.socket.setsockopt(zmq.IPV6, 1)
            elif hasattr(zmq, "IPV4ONLY"):
                self.socket.setsockopt(zmq.IPV4ONLY, 0)
        self.socket.linger = self.linger
        self.socket.connect(self.master_uri)

    # TODO: timeout all in-flight sessions, or error
    def close(self):
        if self._closing:
            return
        self._closing = True
        if self.socket:
            self.socket.close()
            self.socket = None
        if self.context and self.context.closed is False:
            # This hangs if closing the stream causes an import error
            self.context.term()
            self.context = None

    async def _send_recv(self, message):
        message = salt.payload.dumps(message)
        async with self.sending:
            try:
                await self.socket.send(message)
                ret = await self.socket.recv()
            except zmq.error.ZMQError:
                self.close()
                await self.connect()
                await self.socket.send(message)
                ret = await self.socket.recv()
        return salt.payload.loads(ret)

    async def send(self, load, timeout=60):
        """
        Return a future which will be completed when the message has a response
        """
        if not self.socket:
            await self.connect()
        try:
            return await asyncio.wait_for(self._send_recv(load), timeout=timeout)
        except (asyncio.exceptions.TimeoutError, TimeoutError):
            self.close()
            raise SaltReqTimeoutError("Request client send timedout")
        except Exception:
            self.close()
            raise

    @staticmethod
    def get_master_uri(opts):
        if "master_uri" in opts:
            return opts["master_uri"]
        if "master_ip" in opts:
            return _get_master_uri(
                opts["master_ip"],
                opts["master_port"],
                source_ip=opts.get("source_ip"),
                source_port=opts.get("source_ret_port"),
            )
        # if we've reached here something is very abnormal
        raise SaltException("ReqChannel: missing master_uri/master_ip in self.opts")



"""
Cobbler Tops
============

Cobbler Tops is a master tops subsystem used to look up mapping information
from Cobbler via its API. The same cobbler.* parameters are used for both
the Cobbler tops and Cobbler pillar modules.

.. code-block:: yaml

  master_tops:
    cobbler: {}
  cobbler.url: https://example.com/cobbler_api #default is http://localhost/cobbler_api
  cobbler.user: username # default is no username
  cobbler.password: password # default is no password


Module Documentation
====================
"""

import logging
import xmlrpc.client  # nosec

log = logging.getLogger(__name__)


__opts__ = {
    "cobbler.url": "http://localhost/cobbler_api",
    "cobbler.user": None,
    "cobbler.password": None,
}


def top(**kwargs):
    """
    Look up top data in Cobbler for a minion.
    """
    url = __opts__["cobbler.url"]
    user = __opts__["cobbler.user"]
    password = __opts__["cobbler.password"]

    minion_id = kwargs["opts"]["id"]

    log.info("Querying cobbler for information for %r", minion_id)
    try:
        server = xmlrpc.client.Server(url, allow_none=True)
        if user:
            server.login(user, password)
        data = server.get_blended_data(None, minion_id)
    except Exception:  # pylint: disable=broad-except
        log.exception("Could not connect to cobbler.")
        return {}

    return {data["status"]: data["mgmt_classes"]}



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.pyenv as pyenv
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {pyenv: {}}


def test_installed():
    """
    Test to verify that the specified python is installed with pyenv.
    """
    name = "python-2.7.6"

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    with patch.dict(pyenv.__opts__, {"test": True}):
        comt = "python 2.7.6 is set to be installed"
        ret.update({"comment": comt})
        assert pyenv.installed(name) == ret

    with patch.dict(pyenv.__opts__, {"test": False}):
        mock_f = MagicMock(side_effect=[False, False, True])
        mock_fa = MagicMock(side_effect=[False, True])
        mock_str = MagicMock(return_value="2.7.6")
        mock_lst = MagicMock(return_value=["2.7.6"])
        with patch.dict(
            pyenv.__salt__,
            {
                "pyenv.is_installed": mock_f,
                "pyenv.install": mock_fa,
                "pyenv.default": mock_str,
                "pyenv.versions": mock_lst,
            },
        ):
            comt = "pyenv failed to install"
            ret.update({"comment": comt, "result": False})
            assert pyenv.installed(name) == ret

            comt = "Requested python exists."
            ret.update({"comment": comt, "result": True, "default": True})
            assert pyenv.installed(name) == ret

            assert pyenv.installed(name) == ret


def test_absent():
    """
    Test to verify that the specified python is not installed with pyenv.
    """
    name = "python-2.7.6"

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    with patch.dict(pyenv.__opts__, {"test": True}):
        comt = "python 2.7.6 is set to be uninstalled"
        ret.update({"comment": comt})
        assert pyenv.absent(name) == ret

    with patch.dict(pyenv.__opts__, {"test": False}):
        mock_f = MagicMock(side_effect=[False, True])
        mock_t = MagicMock(return_value=True)
        mock_str = MagicMock(return_value="2.7.6")
        mock_lst = MagicMock(return_value=["2.7.6"])
        with patch.dict(
            pyenv.__salt__,
            {
                "pyenv.is_installed": mock_f,
                "pyenv.uninstall_python": mock_t,
                "pyenv.default": mock_str,
                "pyenv.versions": mock_lst,
            },
        ):
            comt = "pyenv not installed, 2.7.6 not either"
            ret.update({"comment": comt, "result": True})
            assert pyenv.absent(name) == ret

            comt = "Successfully removed python"
            ret.update(
                {
                    "comment": comt,
                    "result": True,
                    "default": True,
                    "changes": {"2.7.6": "Uninstalled"},
                }
            )
            assert pyenv.absent(name) == ret


def test_install_pyenv():
    """
    Test to install pyenv if not installed.
    """
    name = "python-2.7.6"

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    with patch.dict(pyenv.__opts__, {"test": True}):
        comt = "pyenv is set to be installed"
        ret.update({"comment": comt})
        assert pyenv.install_pyenv(name) == ret

    with patch.dict(pyenv.__opts__, {"test": False}):
        mock_t = MagicMock(return_value=True)
        mock_str = MagicMock(return_value="2.7.6")
        mock_lst = MagicMock(return_value=["2.7.6"])
        with patch.dict(
            pyenv.__salt__,
            {
                "pyenv.install_python": mock_t,
                "pyenv.default": mock_str,
                "pyenv.versions": mock_lst,
            },
        ):
            comt = "Successfully installed python"
            ret.update(
                {
                    "comment": comt,
                    "result": True,
                    "default": False,
                    "changes": {None: "Installed"},
                }
            )
            assert pyenv.install_pyenv(name) == ret



"""
    Test cases for salt.modules.dig
"""

import pytest

import salt.modules.dig as dig
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {dig: {}}


class SpfValues:
    def __call__(self, key, python_shell=False):
        _spf_values = {
            "dig +short xmission.com TXT": {
                "pid": 27282,
                "retcode": 0,
                "stderr": "",
                "stdout": '"v=spf1 a mx include:_spf.xmission.com ?all"',
            },
            "dig +short _spf.xmission.com TXT": {
                "pid": 27282,
                "retcode": 0,
                "stderr": "",
                "stdout": '"v=spf1 a mx ip4:198.60.22.0/24 ip4:166.70.13.0/24 ~all"',
            },
            "dig +short xmission-redirect.com TXT": {
                "pid": 27282,
                "retcode": 0,
                "stderr": "",
                "stdout": "v=spf1 redirect=_spf.xmission.com",
            },
            "dig +short foo.com TXT": {
                "pid": 27282,
                "retcode": 0,
                "stderr": "",
                "stdout": "v=spf1 ip4:216.73.93.70/31 ip4:216.73.93.72/31 ~all",
            },
        }
        return _spf_values.get(
            " ".join(key), {"pid": 27310, "retcode": 0, "stderr": "", "stdout": ""}
        )


def test_dig_cname_found():
    dig_mock = MagicMock(
        return_value={
            "pid": 2018,
            "retcode": 0,
            "stderr": "",
            "stdout": "bellanotte1986.github.io.",
        }
    )
    with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
        assert dig.CNAME("www.eitr.tech") == "bellanotte1986.github.io."


def test_dig_cname_none_found():
    dig_mock = MagicMock(
        return_value={
            "pid": 2022,
            "retcode": 0,
            "stderr": "",
            "stdout": "",
        }
    )
    with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
        assert dig.CNAME("www.google.com") == ""


def test_check_ip():
    assert dig.check_ip("127.0.0.1")


def test_check_ip_ipv6():
    assert dig.check_ip("1111:2222:3333:4444:5555:6666:7777:8888")


def test_check_ip_ipv6_valid():
    assert dig.check_ip("2607:fa18:0:3::4")


def test_check_ip_neg():
    assert not dig.check_ip("-127.0.0.1")


def test_check_ip_empty():
    assert not dig.check_ip("")


def test_a():
    dig_mock = MagicMock(
        return_value={
            "pid": 3656,
            "retcode": 0,
            "stderr": "",
            "stdout": (
                "74.125.193.104\n"
                "74.125.193.105\n"
                "74.125.193.99\n"
                "74.125.193.106\n"
                "74.125.193.103\n"
                "74.125.193.147"
            ),
        }
    )
    with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
        assert dig.A("www.google.com") == [
            "74.125.193.104",
            "74.125.193.105",
            "74.125.193.99",
            "74.125.193.106",
            "74.125.193.103",
            "74.125.193.147",
        ]


def test_ptr():
    dig_mock = MagicMock(
        return_value={
            "pid": 3657,
            "retcode": 0,
            "stderr": "",
            "stdout": ("dns.google."),
        }
    )
    with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
        assert dig.ptr("8.8.8.8") == [
            "dns.google.",
        ]


def test_aaaa():
    dig_mock = MagicMock(
        return_value={
            "pid": 25451,
            "retcode": 0,
            "stderr": "",
            "stdout": "2607:f8b0:400f:801::1014",
        }
    )
    with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
        assert dig.AAAA("www.google.com") == ["2607:f8b0:400f:801::1014"]


def test_ns():
    with patch("salt.modules.dig.A", MagicMock(return_value=["ns4.google.com."])):
        dig_mock = MagicMock(
            return_value={
                "pid": 26136,
                "retcode": 0,
                "stderr": "",
                "stdout": "ns4.google.com.",
            }
        )
        with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
            assert dig.NS("google.com") == ["ns4.google.com."]


def test_spf():
    dig_mock = MagicMock(side_effect=SpfValues())
    with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
        assert dig.SPF("foo.com") == ["216.73.93.70/31", "216.73.93.72/31"]


def test_spf_redir():
    """
    Test for SPF records which use the 'redirect' SPF mechanism
    https://en.wikipedia.org/wiki/Sender_Policy_Framework#Mechanisms
    """
    dig_mock = MagicMock(side_effect=SpfValues())
    with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
        assert dig.SPF("xmission-redirect.com") == ["198.60.22.0/24", "166.70.13.0/24"]


def test_spf_include():
    """
    Test for SPF records which use the 'include' SPF mechanism
    https://en.wikipedia.org/wiki/Sender_Policy_Framework#Mechanisms
    """
    dig_mock = MagicMock(side_effect=SpfValues())
    with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
        assert dig.SPF("xmission.com") == ["198.60.22.0/24", "166.70.13.0/24"]


def test_mx():
    dig_mock = MagicMock(
        return_value={
            "pid": 27780,
            "retcode": 0,
            "stderr": "",
            "stdout": (
                "10 aspmx.l.google.com.\n"
                "20 alt1.aspmx.l.google.com.\n"
                "40 alt3.aspmx.l.google.com.\n"
                "50 alt4.aspmx.l.google.com.\n"
                "30 alt2.aspmx.l.google.com."
            ),
        }
    )
    with patch.dict(dig.__salt__, {"cmd.run_all": dig_mock}):
        assert dig.MX("google.com") == [
            ["10", "aspmx.l.google.com."],
            ["20", "alt1.aspmx.l.google.com."],
            ["40", "alt3.aspmx.l.google.com."],
            ["50", "alt4.aspmx.l.google.com."],
            ["30", "alt2.aspmx.l.google.com."],
        ]



import pytest

import salt.utils.win_pwsh as win_pwsh
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


def test_run_dict():
    """
    Tests the run_dict function
    """
    result = win_pwsh.run_dict("Get-Item C:\\Windows")
    assert result["Name"] == "Windows"
    assert result["FullName"] == "C:\\Windows"


def test_run_dict_json_string():
    """
    Tests the run_dict function with json string
    """
    ret = {
        "pid": 1,
        "retcode": 0,
        "stderr": "",
        "stdout": '[{"HotFixID": "KB123456"}, {"HotFixID": "KB123457"}]',
    }
    mock_all = MagicMock(return_value=ret)
    with patch("salt.modules.cmdmod.run_all", mock_all):
        result = win_pwsh.run_dict("Junk-Command")
        assert result == [{"HotFixID": "KB123456"}, {"HotFixID": "KB123457"}]


def test_run_dict_empty_return():
    """
    Tests the run_dict function with json string
    """
    ret = {
        "pid": 1,
        "retcode": 0,
        "stderr": "",
        "stdout": "",
    }
    mock_all = MagicMock(return_value=ret)
    with patch("salt.modules.cmdmod.run_all", mock_all):
        result = win_pwsh.run_dict("Junk-Command")
        assert result == {}


def test_run_dict_stderr():
    ret = {
        "pid": 1,
        "retcode": 1,
        "stderr": "This is an error",
        "stdout": "",
    }
    mock_all = MagicMock(return_value=ret)
    with patch("salt.modules.cmdmod.run_all", mock_all):
        with pytest.raises(CommandExecutionError) as exc_info:
            win_pwsh.run_dict("Junk-Command")
        assert "This is an error" in exc_info.value.message


def test_run_dict_missing_retcode():
    ret = {
        "pid": 1,
        "stderr": "",
        "stdout": "",
    }
    mock_all = MagicMock(return_value=ret)
    with patch("salt.modules.cmdmod.run_all", mock_all):
        with pytest.raises(CommandExecutionError) as exc_info:
            win_pwsh.run_dict("Junk-Command")
        assert "Issue executing PowerShell" in exc_info.value.message


def test_run_dict_retcode_not_zero():
    ret = {
        "pid": 1,
        "retcode": 1,
        "stderr": "",
        "stdout": "",
    }
    mock_all = MagicMock(return_value=ret)
    with patch("salt.modules.cmdmod.run_all", mock_all):
        with pytest.raises(CommandExecutionError) as exc_info:
            win_pwsh.run_dict("Junk-Command")
        assert "Issue executing PowerShell" in exc_info.value.message


def test_run_dict_invalid_json():
    ret = {
        "pid": 1,
        "retcode": 0,
        "stderr": "",
        "stdout": "Invalid Json",
    }
    mock_all = MagicMock(return_value=ret)
    with patch("salt.modules.cmdmod.run_all", mock_all):
        with pytest.raises(CommandExecutionError) as exc_info:
            win_pwsh.run_dict("Junk-Command")
        assert "No JSON results from PowerShell" in exc_info.value.message
